# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2003-2005  Donald N. Allingham
# Copyright (C) 2023       Avi Markovitz
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
"""
Hebrew-specific classes for relationships.
"""
# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import logging

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
import gramps.gen.relationship
from gramps.gen.lib import Person

MALE = Person.MALE
FEMALE = Person.FEMALE
UNKNOWN = Person.UNKNOWN

LOG = logging.getLogger("gen.relationship")
LOG.addHandler(logging.StreamHandler())

# -------------------------------------------------------------------------
#
#
#
# -------------------------------------------------------------------------

_LEVEL_NAME = [
    "",
    "מרמה ראשונה",
    "מרמה שניה",
    "מרמה שלישית",
    "מרמה רביעית",
    "מרמה חמישית",
    "מרמה שישית",
    "מרמה שביעית",
    "מרמה שמינית",
    "מרמה תשיעית",
    "מרמה עשירית",
    "מרמה אחת עשרה",
]

_REMOVED_LEVEL = [
    "",
    "מדרגה שניה",
    "פעמיים מדרגה שניה",
    "שלוש פעמים מדרגה שניה",
    "ארבע פעמים מדרגה שניה",
    "חמש פעמים מדרגה שניה",
    "שש פעמים מדרגה שניה",
    "שבע פעמים מדרגה שניה",
    "שמונה פעמים מדרגה שניה",
    "תשע פעמים מדרגה שניה",
    "עשר פעמים מדרגה שניה",
    "אחת עשרה פעמים מדרגה שניה",
]

_PARENTS_LEVEL = [
    "",
    "הורים",
    "סבים",
    "סבות",
    "סב גדול",
    "סבה גדולה",
    "סב וסבה גדולים",
    "סב וסבה מרמה שלישית",
    "סב וסבה מרמה רביעית",
    "סב וסבה מרמה חמישית",
    "סב וסבה מרמה שישית",
    "סב וסבה מרמה שביעית",
    "סב וסבה מרמה שמינית",
    "סב וסבה מרמה תשיעית",
    "סב וסבה מרמה עשירית",
    "סב וסבה מרמה אחת עשרה",
]

_FATHER_LEVEL = [
    "",
    "סב %(step)s %(inlaw)s",
    "סב גדול %(step)s %(inlaw)s",
    "סב מרמה שלישית %(step)s %(inlaw)s",
    "סב מרמה רביעית %(step)s %(inlaw)s",
    "סב מרמה חמישית %(step)s %(inlaw)s",
    "סב מרמה שישית %(step)s %(inlaw)s",
    "סב מרמה שביעית %(step)s %(inlaw)s",
    "סב מרמה שמינית %(step)s %(inlaw)s",
    "סב מרמה תשיעית %(step)s %(inlaw)s",
    "סב מרמה עשירית %(step)s %(inlaw)s",
    "סב מרמה אחת עשרה %(step)s %(inlaw)s",
]

_MOTHER_LEVEL = [
    "",
    "סבה %(step)s %(inlaw)s",
    "סבה גדולה %(step)s %(inlaw)s",
    "סבה מרמה שלישית %(step)s %(inlaw)s",
    "סבה מרמה רביעית %(step)s %(inlaw)s",
    "סבה מרמה חמישית %(step)s %(inlaw)s",
    "סבה מרמה שישית %(step)s %(inlaw)s",
    "סבה מרמה שביעית %(step)s %(inlaw)s",
    "סבה מרמה שמינית %(step)s %(inlaw)s",
    "סבה מרמה תשיעית %(step)s %(inlaw)s",
    "סבה מרמה עשירית %(step)s %(inlaw)s",
    "סבה מרמה אחת עשרה %(step)s %(inlaw)s",
]

_SON_LEVEL = [
    "",
    "בן %(step)s %(inlaw)s",
    "נכד %(step)s %(inlaw)s",
    "נין %(step)s %(inlaw)s",
    "חימש %(step)s %(inlaw)s",
    "נכד מרמה חמישית %(step)s %(inlaw)s",
    "נכד מרמה שישית %(step)s %(inlaw)s",
    "נכד מרמה שביעית %(step)s %(inlaw)s",
    "נכד מרמה שמינית %(step)s %(inlaw)s",
    "נכד מרמה תשיעית %(step)s %(inlaw)s",
    "נכד מרמה עשירית %(step)s %(inlaw)s",
    "נכד מרמה אחת עשרה %(step)s %(inlaw)s",
    "נכד רחוק %(step)s %(inlaw)s",
    "נכד רחוק %(step)s %(inlaw)s",
    "נכד רחוק %(step)s %(inlaw)s",
    "נכד רחוק %(step)s %(inlaw)s",
    "נכד רחוק %(step)s %(inlaw)s",
    "נכד רחוק %(step)s %(inlaw)s",
    "נכד רחוק %(step)s %(inlaw)s",
    "נכד רחוק %(step)s %(inlaw)s",
    "נכד רחוק %(step)s %(inlaw)s",
    "נכד רחוק %(step)s %(inlaw)s",
    "נכד רחוק %(step)s %(inlaw)s",
    "נכד רחוק %(step)s %(inlaw)s",
    "נכד רחוק %(step)s %(inlaw)s",
    "נכד רחוק %(step)s %(inlaw)s",
    "נכד רחוק %(step)s %(inlaw)s",
    "נכד רחוק %(step)s %(inlaw)s",
    "נכד רחוק %(step)s %(inlaw)s",
    "נכד רחוק %(step)s %(inlaw)s",
]

_DAUGHTER_LEVEL = [
    "",
    "בת %(step)s %(inlaw)s",
    "נכדה %(step)s %(inlaw)s",
    "נינה %(step)s %(inlaw)s",
    "חימשה %(step)s %(inlaw)s",
    "נכדה מרמה חמישית %(step)s %(inlaw)s",
    "נכדה מרמה שישית %(step)s %(inlaw)s",
    "נכדה מרמה שביעית %(step)s %(inlaw)s",
    "נכדה מרמה שמינית %(step)s %(inlaw)s",
    "נכדה מרמה תשיעית %(step)s %(inlaw)s",
    "נכדה מרמה עשירית %(step)s %(inlaw)s",
    "נכדה מרמה אחת עשרה %(step)s %(inlaw)s",
    "נכדה רחוקה %(step)s %(inlaw)s",
    "נכדה רחוקה %(step)s %(inlaw)s",
    "נכדה רחוקה %(step)s %(inlaw)s",
    "נכדה רחוקה %(step)s %(inlaw)s",
    "נכדה רחוקה %(step)s %(inlaw)s",
    "נכדה רחוקה %(step)s %(inlaw)s",
    "נכדה רחוקה %(step)s %(inlaw)s",
    "נכדה רחוקה %(step)s %(inlaw)s",
    "נכדה רחוקה %(step)s %(inlaw)s",
    "נכדה רחוקה %(step)s %(inlaw)s",
    "נכדה רחוקה %(step)s %(inlaw)s",
    "נכדה רחוקה %(step)s %(inlaw)s",
    "נכדה רחוקה %(step)s %(inlaw)s",
    "נכדה רחוקה %(step)s %(inlaw)s",
    "נכדה רחוקה %(step)s %(inlaw)s",
    "נכדה רחוקה %(step)s %(inlaw)s",
    "נכדה רחוקה %(step)s %(inlaw)s",
    "נכדה רחוקה %(step)s %(inlaw)s",
]

_SISTER_LEVEL = [
    "",
    "אחות %(step)s %(inlaw)s",
    "דודה %(step)s %(inlaw)s",
    "דודה גדולה %(step)s %(inlaw)s",
    "דודה מרמה שלישית %(step)s %(inlaw)s",
    "דודה מרמה רביעית %(step)s %(inlaw)s",
    "דודה מרמה חמישית %(step)s %(inlaw)s",
    "דודה מרמה שישית %(step)s %(inlaw)s",
    "דודה מרמה שביעית %(step)s %(inlaw)s",
    "דודה מרמה שמינית %(step)s %(inlaw)s",
    "דודה מרמה תשיעיתית %(step)s %(inlaw)s",
    "דודה מרמה עשירית %(step)s %(inlaw)s",
    "דודה מרמה אחת עשרה%(step)s %(inlaw)s",
]

_BROTHER_LEVEL = [
    "",
    "אח %(step)s %(inlaw)s",
    "דוד %(step)s %(inlaw)s",
    "דוד גדול %(step)s %(inlaw)s",
    "דוד מרמה שלישית %(step)s %(inlaw)s",
    "דוד מרמה רביעית %(step)s %(inlaw)s",
    "דוד מרמה חמישית %(step)s %(inlaw)s",
    "דוד מרמה שישית %(step)s %(inlaw)s",
    "דוד מרמה שביעית %(step)s %(inlaw)s",
    "דוד מרמה שמינית %(step)s %(inlaw)s",
    "דוד מרמה תשיעית %(step)s %(inlaw)s",
    "דוד מרמה עשירית %(step)s %(inlaw)s",
    "דוד מרמה אחת עשרה %(step)s %(inlaw)s",
]

_NEPHEW_LEVEL = [
    "",
    "אחיין %(step)s %(inlaw)s",
    "נכדן %(step)s %(inlaw)s",
    "אחיין גדול %(step)s %(inlaw)s",
    "אחיין מרמה שלישית %(step)s %(inlaw)s",
    "אחיין מרמה רביעית %(step)s %(inlaw)s",
    "אחיין מרמה חמישית %(step)s %(inlaw)s",
    "אחיין מרמה שישית %(step)s %(inlaw)s",
    "אחיין מרמה שביעית %(step)s %(inlaw)s",
    "אחיין מרמה שמינית %(step)s %(inlaw)s",
    "אחיין מרמה תשיעית %(step)s %(inlaw)s",
    "אחיין מרמה עשירית %(step)s %(inlaw)s",
    "אחיין מרמה אחת עשרה %(step)s %(inlaw)s",
]

_NIECE_LEVEL = [
    "",
    "אחיינית %(step)s %(inlaw)s",
    "נכדנית %(step)s %(inlaw)s",
    "אחיינית גדולה %(step)s %(inlaw)s",
    "אחיינית מרמה שלישית %(step)s %(inlaw)s",
    "אחיינית מרמה רביעית %(step)s %(inlaw)s",
    "אחיינית מרמה חמישית %(step)s %(inlaw)s",
    "אחיינית מרמה שישית %(step)s %(inlaw)s",
    "אחיינית מרמה שביעית %(step)s %(inlaw)s",
    "אחיינית מרמה שמינית %(step)s %(inlaw)s",
    "אחיינית מרמה תשיעית %(step)s %(inlaw)s",
    "אחיינית מרמה עשירית %(step)s %(inlaw)s",
    "אחיינית מרמה אחת עשרה %(step)s %(inlaw)s",
]

_CHILDREN_LEVEL = [
    "",
    "ילדים",
    "נכדים",
    "נינים",
    "חימשים",
    "נכדים מרמה חמישית",
    "נכדים מרמה שישית",
    "נכדים מרמה שביעית",
    "נכדים מרמה שמינית",
    "נכדים מרמה תשיעית",
    "נכדים מרמה עשירית",
    "נכדים מרמה אחת עשרה",
]

_SIBLINGS_LEVEL = [
    "",
    "אחאים",
    "דוד/דודה",
    "דוד/דודה גדולים",
    "דוד/דודה מרמה שלישית",
    "דוד/דודה מרמה רביעית",
    "דוד/דודה מרמה חמישית",
    "דוד/דודה מרמה שישית",
    "דוד/דודה מרמה שביעית",
    "דוד/דודה מרמה שמינית",
    "דוד/דודה מרמה תשיעית",
    "דוד/דודה מרמה עשירית",
    "דוד/דודה מרמה אחת עשרה",
]

_SIBLING_LEVEL = [
    "",
    "אחאים %(step)s %(inlaw)s",
    "דוד/דודה %(step)s %(inlaw)s",
    "דוד/דודה גדולים %(step)s %(inlaw)s",
    "דוד/דודה מרמה שלישית %(step)s %(inlaw)s",
    "דוד/דודה מרמה רביעית %(step)s %(inlaw)s",
    "דוד/דודה מרמה חמישית %(step)s %(inlaw)s",
    "דוד/דודה מרמה שישית %(step)s %(inlaw)s",
    "דוד/דודה מרמה שביעית %(step)s %(inlaw)s",
    "דוד/דודה מרמה שמינית %(step)s %(inlaw)s",
    "דוד/דודה מרמה תשיעית %(step)s %(inlaw)s",
    "דוד/דודה מרמה עשירית %(step)s %(inlaw)s",
    "דוד/דודה מרמה אחת עשרה %(step)s %(inlaw)s",
]

_NEPHEWS_NIECES_LEVEL = [
    "",
    "אחאים",
    "אחיין/אחיינית",
    "נכדן/נכדנית",
    "אחיין/אחיינית גדולים",
    "אחיין/אחיינית שלישית",
    "אחיין/אחיינית רביעית",
    "אחיין/אחיינית חמישית",
    "אחיין/אחיינית מרמה שישית",
    "אחיין/אחיינית מרמה שביעית",
    "אחיין/אחיינית מרמה שמינית",
    "אחיין/אחיינית מרמה תשיעית",
    "אחיין/אחיינית מרמה עשירית",
    "אחיין/אחיינית מרמה אחת עשרה",
]


# -------------------------------------------------------------------------
#
# RelationshipCalculator
#
# -------------------------------------------------------------------------
class RelationshipCalculator(gramps.gen.relationship.RelationshipCalculator):
    """
    The relationship calculator helps to determine the relationship between
    two people.
    """

    # sibling strings  for Hebrew we need four "step": male sing/plur, female sing/plur
    STEP = "שלוב"
    STEP_F = "שלובה"
    STEP_M = "שלוב"  # this is actually redundant if Can't make it "plural form".
    HALF = "למחצה"
    INLAW = "מחיתון"

    def __init__(self):
        gramps.gen.relationship.RelationshipCalculator.__init__(self)

    DIST_FATHER = "אב־קדמון רחוק %(step)s %(inlaw)s (%(level)d דורות)"

    def _get_father(self, level, step="", inlaw=""):
        """
        Internal english method to create relation string
        """
        if level > len(_FATHER_LEVEL) - 1:
            return self.DIST_FATHER % {"step": step, "inlaw": inlaw, "level": level}
        else:
            return _FATHER_LEVEL[level] % {"step": step, "inlaw": inlaw}

    DIST_SON = "בן רחוק %(step) %(inlaw)s (%(level)d דורות)"

    def _get_son(self, level, step="", inlaw=""):
        """
        Internal english method to create relation string
        """
        if level > len(_SON_LEVEL) - 1:
            return self.DIST_SON % {"step": step, "inlaw": inlaw, "level": level}
        else:
            return _SON_LEVEL[level] % {"step": step, "inlaw": inlaw}

    DIST_MOTHER = "אם־קדמונית רחוקה %(step)s %(inlaw) s(%(level)d דורות)"

    def _get_mother(self, level, step="", inlaw=""):
        """
        Internal english method to create relation string
        """
        if level > len(_MOTHER_LEVEL) - 1:
            return self.DIST_MOTHER % {"step": step, "inlaw": inlaw, "level": level}
        else:
            return _MOTHER_LEVEL[level] % {"step": step, "inlaw": inlaw}

    DIST_DAUGHTER = "בת רחוקה %(step) %(inlaw)s(%(level)d דורות)"

    def _get_daughter(self, level, step="", inlaw=""):
        """
        Internal english method to create relation string
        """
        if level > len(_DAUGHTER_LEVEL) - 1:
            return self.DIST_DAUGHTER % {"step": step, "inlaw": inlaw, "level": level}
        else:
            return _DAUGHTER_LEVEL[level] % {"step": step, "inlaw": inlaw}

    def _get_parent_unknown(self, level, step="", inlaw=""):
        """
        Internal english method to create relation string
        """
        if level < len(_LEVEL_NAME):
            return (
                "אב־קדמון %(step)s %(inlaw)s" % {"step": step, "inlaw": inlaw}
                + _LEVEL_NAME[level]
            )
        else:
            return "אב־קדמון רחוק %s %s (%d דורות)" % (step, inlaw, level)

    DIST_CHILD = "צאצא רחוק %(step)s (%(level)d דורות)"

    def _get_child_unknown(self, level, step="", inlaw=""):
        """
        Internal english method to create relation string
        """
        if level < len(_LEVEL_NAME):
            return (
                "צאצא %(step)s %(inlaw)s" % {"step": step, "inlaw": inlaw}
                + _LEVEL_NAME[level]
            )
        else:
            return self.DIST_CHILD % {"step": step, "level": level}

    DIST_AUNT = "דודה רחוקה %(step)s %(inlaw)s"

    def _get_aunt(self, level, step="", inlaw=""):
        """
        Internal english method to create relation string
        """
        if level > len(_SISTER_LEVEL) - 1:
            return self.DIST_AUNT % {"step": step, "inlaw": inlaw}
        else:
            return _SISTER_LEVEL[level] % {"step": step, "inlaw": inlaw}

    DIST_UNCLE = "דוד רחוק %(step)s %(inlaw)s"

    def _get_uncle(self, level, step="", inlaw=""):
        """
        Internal english method to create relation string
        """
        if level > len(_BROTHER_LEVEL) - 1:
            return self.DIST_UNCLE % {"step": step, "inlaw": inlaw}
        else:
            return _BROTHER_LEVEL[level] % {"step": step, "inlaw": inlaw}

    DIST_NEPHEW = "אחיין רחוק %(step)s %(inlaw)s"

    def _get_nephew(self, level, step="", inlaw=""):
        """
        Internal english method to create relation string
        """
        if level > len(_NEPHEW_LEVEL) - 1:
            return self.DIST_NEPHEW % {"step": step, "inlaw": inlaw}
        else:
            return _NEPHEW_LEVEL[level] % {"step": step, "inlaw": inlaw}

    DIST_NIECE = "אחיינית רחוקה %(step)s %(inlaw)s"

    def _get_niece(self, level, step="", inlaw=""):
        """
        Internal english method to create relation string
        """
        if level > len(_NIECE_LEVEL) - 1:
            return self.DIST_NIECE % {"step": step, "inlaw": inlaw}
        else:
            return _NIECE_LEVEL[level] % {"step": step, "inlaw": inlaw}

    def _get_cousin(self, level, removed, dir="", step="", inlaw=""):
        """
        Internal english method to create relation string
        """
        if removed == 0 and level < len(_LEVEL_NAME):
            return "בן־דוד %s %s %s" % (step, inlaw, _LEVEL_NAME[level])
        elif removed > len(_REMOVED_LEVEL) - 1 or level > len(_LEVEL_NAME) - 1:
            return "קרוב־משפחה רחוק %s %s" % (step, inlaw)
        else:
            return "בן־דוד/בת־דודה %s %s %s %s %s" % (
                step,
                inlaw,
                _LEVEL_NAME[level],
                _REMOVED_LEVEL[removed],
                dir,
            )

    DIST_SIB = "דוד/דודה רחוקים %(step)s %(inlaw)s"

    def _get_sibling(self, level, step="", inlaw=""):
        """
        Internal english method to create relation string
        """
        if level < len(_SIBLING_LEVEL):
            return _SIBLING_LEVEL[level] % {"step": step, "inlaw": inlaw}
        else:
            return self.DIST_SIB % {"step": step, "inlaw": inlaw}

    def get_plural_relationship_string(
        self,
        Ga,
        Gb,
        reltocommon_a="",
        reltocommon_b="",
        only_birth=True,
        in_law_a=False,
        in_law_b=False,
    ):
        """
        Provide a string that describes the relationsip between a person, and
        a group of people with the same relationship. E.g. "grandparents" or
        "children".

        Ga and Gb can be used to mathematically calculate the relationship.

        .. seealso::
            http://en.wikipedia.org/wiki/Cousin#Mathematical_definitions

        :param Ga: The number of generations between the main person and the
                   common ancestor.
        :type Ga: int
        :param Gb: The number of generations between the group of people and the
                   common ancestor
        :type Gb: int
        :param reltocommon_a: relation path to common ancestor or common
                              Family for person a.
                              Note that length = Ga
        :type reltocommon_a: str
        :param reltocommon_b: relation path to common ancestor or common
                              Family for person b.
                              Note that length = Gb
        :type reltocommon_b: str
        :param only_birth: True if relation between a and b is by birth only
                           False otherwise
        :type only_birth: bool
        :param in_law_a: True if path to common ancestors is via the partner
                         of person a
        :type in_law_a: bool
        :param in_law_b: True if path to common ancestors is via the partner
                         of person b
        :type in_law_b: bool
        :returns: A string describing the relationship between the person and
                  the group.
        :rtype: str
        """
        rel_str = "קרובי־משפחה רחוקים"
        if Ga == 0:
            # These are descendants
            if Gb < len(_CHILDREN_LEVEL):
                rel_str = _CHILDREN_LEVEL[Gb]
            else:
                rel_str = "צאצאים רחוקים"
        elif Gb == 0:
            # These are parents/grand parents
            if Ga < len(_PARENTS_LEVEL):
                rel_str = _PARENTS_LEVEL[Ga]
            else:
                rel_str = "אבות־קדמונים רחוקים"
        elif Gb == 1:
            # These are siblings/aunts/uncles
            if Ga < len(_SIBLINGS_LEVEL):
                rel_str = _SIBLINGS_LEVEL[Ga]
            else:
                rel_str = "דודים/דודות רחוקים"
        elif Ga == 1:
            # These are nieces/nephews
            if Gb < len(_NEPHEWS_NIECES_LEVEL):
                rel_str = _NEPHEWS_NIECES_LEVEL[Gb]
            else:
                rel_str = "אחיינים/אחייניות רחוקים"
        elif Ga > 1 and Ga == Gb:
            # These are cousins in the same generation
            if Ga <= len(_LEVEL_NAME):
                rel_str = "בני דודים %s  " % _LEVEL_NAME[Ga - 1]
            else:
                rel_str = "בני דודים רחוקים"
        elif Ga > 1 and Ga > Gb:
            # These are cousins in different generations with the second person
            # being in a higher generation from the common ancestor than the
            # first person.
            if Gb <= len(_LEVEL_NAME) and (Ga - Gb) < len(_REMOVED_LEVEL):
                rel_str = " %s %s (עולה)" % (
                    _LEVEL_NAME[Gb - 1],
                    _REMOVED_LEVEL[Ga - Gb],
                )
            else:
                rel_str = "בני דודים רחוקים"
        elif Gb > 1 and Gb > Ga:
            # These are cousins in different generations with the second person
            # being in a lower generation from the common ancestor than the
            # first person.
            if Ga <= len(_LEVEL_NAME) and (Gb - Ga) < len(_REMOVED_LEVEL):
                rel_str = " בני דודים%s %s (יורד)" % (
                    _LEVEL_NAME[Ga - 1],
                    _REMOVED_LEVEL[Gb - Ga],
                )
            else:
                rel_str = "בני דודים רחוקים"

        if in_law_b is True:
            rel_str = "זוג של %s" % rel_str

        return rel_str.strip()

    def get_single_relationship_string(
        self,
        Ga,
        Gb,
        gender_a,
        gender_b,
        reltocommon_a,
        reltocommon_b,
        only_birth=True,
        in_law_a=False,
        in_law_b=False,
    ):
        """
        Provide a string that describes the relationsip between a person, and
        another person. E.g. "grandparent" or "child".

        To be used as: 'person b is the grandparent of a', this will be in
        translation string:  'person b is the %(relation)s of a'

        Note that languages with gender should add 'the' inside the
        translation, so eg in french:  'person b est %(relation)s de a'
        where relation will be here: le grandparent

        Ga and Gb can be used to mathematically calculate the relationship.

        .. seealso::
            http://en.wikipedia.org/wiki/Cousin#Mathematical_definitions

        Some languages need to know the specific path to the common ancestor.
        Those languages should use reltocommon_a and reltocommon_b which is
        a string like 'mfmf'.

        The possible string codes are:

        =======================  ===========================================
        Code                     Description
        =======================  ===========================================
        REL_MOTHER               # going up to mother
        REL_FATHER               # going up to father
        REL_MOTHER_NOTBIRTH      # going up to mother, not birth relation
        REL_FATHER_NOTBIRTH      # going up to father, not birth relation
        REL_FAM_BIRTH            # going up to family (mother and father)
        REL_FAM_NONBIRTH         # going up to family, not birth relation
        REL_FAM_BIRTH_MOTH_ONLY  # going up to fam, only birth rel to mother
        REL_FAM_BIRTH_FATH_ONLY  # going up to fam, only birth rel to father
        =======================  ===========================================

        Prefix codes are stripped, so REL_FAM_INLAW_PREFIX is not present.
        If the relation starts with the inlaw of the person a, then 'in_law_a'
        is True, if it starts with the inlaw of person b, then 'in_law_b' is
        True.

        Also REL_SIBLING (# going sideways to sibling (no parents)) is not
        passed to this routine. The collapse_relations changes this to a
        family relation.

        Hence, calling routines should always strip REL_SIBLING and
        REL_FAM_INLAW_PREFIX before calling get_single_relationship_string()
        Note that only_birth=False, means that in the reltocommon one of the
        NOTBIRTH specifiers is present.

        The REL_FAM identifiers mean that the relation is not via a common
        ancestor, but via a common family (note that that is not possible for
        direct descendants or direct ancestors!). If the relation to one of the
        parents in that common family is by birth, then 'only_birth' is not
        set to False. The only_birth() method is normally used for this.

        :param Ga: The number of generations between the main person and the
                   common ancestor.
        :type Ga: int
        :param Gb: The number of generations between the other person and the
                   common ancestor.
        :type Gb: int
        :param gender_a: gender of person a
        :type gender_a: int gender
        :param gender_b: gender of person b
        :type gender_b: int gender
        :param reltocommon_a: relation path to common ancestor or common
                              Family for person a.
                              Note that length = Ga
        :type reltocommon_a: str
        :param reltocommon_b: relation path to common ancestor or common
                              Family for person b.
                              Note that length = Gb
        :type reltocommon_b: str
        :param in_law_a:  True if path to common ancestors is via the partner
                          of person a
        :type in_law_a: bool
        :param in_law_b: True if path to common ancestors is via the partner
                         of person b
        :type in_law_b: bool
        :param only_birth: True if relation between a and b is by birth only
                           False otherwise
        :type only_birth: bool
        :returns: A string describing the relationship between the two people
        :rtype: str

        .. note:: 1. the self.REL_SIBLING should not be passed to this routine,
                     so we should not check on it. All other self.
                  2. for better determination of siblings, use if Ga=1=Gb
                     get_sibling_relationship_string
        """
        if only_birth:
            step = ""
        else:
            if gender_b == MALE:
                step = self.STEP
            elif gender_b == FEMALE:
                step = self.STEP_F
            else:
                step = (
                    self.STEP
                )  # Change this as appropriate for other and unknown gender

        if in_law_a or in_law_b:
            inlaw = self.INLAW
        else:
            inlaw = ""

        rel_str = "קרוב־משפחה רחוק %s%s" % (step, inlaw)

        if Ga == 0:
            # b is descendant of a
            if Gb == 0:
                rel_str = "האדם"
            elif gender_b == MALE:
                rel_str = self._get_son(Gb, step, inlaw)
            elif gender_b == FEMALE:
                rel_str = self._get_daughter(Gb, step, inlaw)
            else:
                rel_str = self._get_child_unknown(Gb, step, inlaw)
        elif Gb == 0:
            # b is parents/grand parent of a
            if gender_b == MALE:
                rel_str = self._get_father(Ga, step, inlaw)
            elif gender_b == FEMALE:
                rel_str = self._get_mother(Ga, step, inlaw)
            else:
                rel_str = self._get_parent_unknown(Ga, step, inlaw)
        elif Gb == 1:
            # b is sibling/aunt/uncle of a
            if gender_b == MALE:
                rel_str = self._get_uncle(Ga, step, inlaw)
            elif gender_b == FEMALE:
                rel_str = self._get_aunt(Ga, step, inlaw)
            else:
                rel_str = self._get_sibling(Ga, step, inlaw)
        elif Ga == 1:
            # b is niece/nephew of a
            if gender_b == MALE:
                rel_str = self._get_nephew(Gb - 1, step, inlaw)
            elif gender_b == FEMALE:
                rel_str = self._get_niece(Gb - 1, step, inlaw)
            elif Gb < len(_NIECE_LEVEL) and Gb < len(_NEPHEW_LEVEL):
                rel_str = "%sאו %s" % (
                    self._get_nephew(Gb - 1, step, inlaw),
                    self._get_niece(Gb - 1, step, inlaw),
                )
            else:
                rel_str = "אחיין/אחיינית רחוקים %s %s" % (step, inlaw)
        elif Ga == Gb:
            # a and b cousins in the same generation
            rel_str = self._get_cousin(Ga - 1, 0, dir="", step=step, inlaw=inlaw)
        elif Ga > Gb:
            # These are cousins in different generations with the second person
            # being in a higher generation from the common ancestor than the
            # first person.
            rel_str = self._get_cousin(
                Gb - 1, Ga - Gb, dir=" (עולה)", step=step, inlaw=inlaw
            )
        elif Gb > Ga:
            # These are cousins in different generations with the second person
            # being in a lower generation from the common ancestor than the
            # first person.
            rel_str = self._get_cousin(
                Ga - 1, Gb - Ga, dir=" (יורד)", step=step, inlaw=inlaw
            )
        return rel_str.strip()

    def get_sibling_relationship_string(
        self, sib_type, gender_a, gender_b, in_law_a=False, in_law_b=False
    ):
        """
        Determine the string giving the relation between two siblings of
        type sib_type.
        Eg: b is the brother of a
        Here 'brother' is the string we need to determine
        This method gives more details about siblings than
        get_single_relationship_string can do.

        .. warning:: DON'T TRANSLATE THIS PROCEDURE IF LOGIC IS EQUAL IN YOUR
                     LANGUAGE, AND SAME METHODS EXIST (get_uncle, get_aunt,
                     get_sibling)
        """
        if sib_type == self.NORM_SIB or sib_type == self.UNKNOWN_SIB:
            typestr = ""
        elif sib_type == self.HALF_SIB_MOTHER or sib_type == self.HALF_SIB_FATHER:
            typestr = self.HALF
        elif sib_type == self.STEP_SIB:
            typestr = self.STEP

        if in_law_a or in_law_b:
            inlaw = self.INLAW
        else:
            inlaw = ""

        if gender_b == MALE:
            rel_str = self._get_uncle(1, typestr, inlaw)
        elif gender_b == FEMALE:
            rel_str = self._get_aunt(1, typestr, inlaw)
        else:
            rel_str = self._get_sibling(1, typestr, inlaw)
        return rel_str.strip()

    def get_partner_relationship_string(self, spouse_type, gender_a, gender_b):
        """
        Determine the string giving the relation between two partners of
        type spouse_type.
        Eg: b is the spouse of a
        Here 'spouse' is the string we need to determine

        .. warning:: DON'T TRANSLATE THIS PROCEDURE IF LOGIC IS EQUAL IN YOUR
                     LANGUAGE, AS GETTEXT IS ALREADY USED !
        """
        # english only needs gender of b, we don't guess if unknown like in old
        # procedure as that is stupid in present day cases!
        gender = gender_b

        if not spouse_type:
            return ""

        if spouse_type == self.PARTNER_MARRIED:
            if gender == MALE:
                return "בעל"
            elif gender == FEMALE:
                return "רעיה"
            else:
                return "זוג"
        elif spouse_type == self.PARTNER_EX_MARRIED:
            if gender == MALE:
                return "בעל לשעבר"
            elif gender == FEMALE:
                return "רעיה לשעבר"
            else:
                return "זוג לשעבר"
        elif spouse_type == self.PARTNER_UNMARRIED:
            if gender == MALE:
                return "שותף"
            elif gender == FEMALE:
                return "שותפה"
            else:
                return "שותף"
        elif spouse_type == self.PARTNER_EX_UNMARRIED:
            if gender == MALE:
                return "שותף לשעבר"
            elif gender == FEMALE:
                return "שותפה לשעבר"
            else:
                return "שותף לשעבר"
        elif spouse_type == self.PARTNER_CIVIL_UNION:
            if gender == MALE:
                return "שותף"
            elif gender == FEMALE:
                return "שותפה"
            else:
                return "שותף"
        elif spouse_type == self.PARTNER_EX_CIVIL_UNION:
            if gender == MALE:
                return "שותף קודם"
            elif gender == FEMALE:
                return "שותפה קודמת"
            else:
                return "שותף קודם"
        elif spouse_type == self.PARTNER_UNKNOWN_REL:
            if gender == MALE:
                return "שותף"
            elif gender == FEMALE:
                return "שותפה"
            else:
                return "שותף"
        else:
            # here we have spouse_type == self.PARTNER_EX_UNKNOWN_REL
            #   or other not catched types
            if gender == MALE:
                return "שותף קודם"
            elif gender == FEMALE:
                return "שותפה קודמת"
            else:
                return "שותף קודם"


if __name__ == "__main__":
    """
    TRANSLATORS, copy this if statement at the bottom of your
    rel_xx.py module, after adding: 'from Relationship import test'
    and test your work with:
    export PYTHONPATH=/path/to/gramps
    python gramps/plugins/rel/rel_xx.py

    See eg rel_fr.py at the bottom
    """
    from gramps.gen.relationship import test

    REL_CALC = RelationshipCalculator()
    test(REL_CALC, True)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
# Copyright (C) 2011       Tim G L Lyons
# Copyright (C) 2012,2024  Doug Blank <doug.blank@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Package providing filtering framework for Gramps.
"""

import logging
import time

# ------------------------------------------------------------------------
#
# Gramps imports
#
# ------------------------------------------------------------------------
from ..lib.person import Person
from ..lib.family import Family
from ..lib.src import Source
from ..lib.citation import Citation
from ..lib.event import Event
from ..lib.place import Place
from ..lib.repo import Repository
from ..lib.media import Media
from ..lib.note import Note
from ..lib.tag import Tag
from ..const import GRAMPS_LOCALE as glocale
from .rules import Rule
from .optimizer import Optimizer

_ = glocale.translation.gettext
LOG = logging.getLogger(".filter.results")


# -------------------------------------------------------------------------
#
# GenericFilter
#
# -------------------------------------------------------------------------
class GenericFilter:
    """Filter class that consists of several rules."""

    logical_functions = ["and", "or", "one"]

    def __init__(self, source=None):
        if source:
            self.need_param = source.need_param
            self.flist = source.flist[:]
            self.name = source.name
            self.comment = source.comment
            self.logical_op = source.logical_op
            self.invert = source.invert
        else:
            self.need_param = 0
            self.flist = []
            self.name = ""
            self.comment = ""
            self.logical_op = "and"
            self.invert = False

    def match(self, handle, db):
        """
        Return True or False depending on whether the handle matches the filter.
        """
        obj = self.get_object(db, handle)
        return self.apply_to_one(db, obj)

    def is_empty(self):
        return (len(self.flist) == 0) or (
            len(self.flist) == 1 and ((self.flist[0].is_empty() and not self.invert))
        )

    def set_logical_op(self, val):
        if val in GenericFilter.logical_functions:
            self.logical_op = val
        else:
            raise Exception("invalid operator: %r" % val)

    def set_invert(self, val):
        self.invert = bool(val)

    def get_invert(self):
        return self.invert

    def get_name(self, ulocale=glocale):
        return self.name

    def set_name(self, name):
        self.name = name

    def set_comment(self, comment):
        self.comment = comment

    def get_comment(self):
        return self.comment

    def add_rule(self, rule):
        self.flist.append(rule)

    def delete_rule(self, rule):
        self.flist.remove(rule)

    def set_rules(self, rules):
        self.flist = rules

    def get_rules(self):
        return self.flist

    def get_cursor(self, db):
        return db.get_person_cursor()

    def get_tree_cursor(self, db):
        return db.get_person_cursor()

    def make_obj(self):
        return Person()

    def find_from_handle(self, db, handle):
        return db.get_person_from_handle(handle)

    def get_number(self, db):
        return db.get_number_of_people()

    def apply_logical_op_to_all(
        self, db, id_list, apply_logical_op, user=None, tupleind=None, tree=False
    ):
        final_list = []

        optimizer = Optimizer(self)
        handles_in, handles_out = optimizer.get_handles()

        LOG.debug(
            "Optimizer handles_in: %s",
            len(handles_in) if handles_in is not None else None,
        )
        LOG.debug("Optimizer handles_out: %s", len(handles_out))
        if id_list is None:
            if handles_in is not None:
                if user:
                    user.begin_progress(_("Filter"), _("Applying ..."), len(handles_in))

                # Use these rather than going through entire database
                for handle in handles_in:
                    if user:
                        user.step_progress()

                    if handle is None:
                        continue

                    obj = self.get_object(db, handle)

                    if apply_logical_op(db, obj, self.flist) != self.invert:
                        final_list.append(obj.handle)

            else:
                with (
                    self.get_tree_cursor(db) if tree else self.get_cursor(db)
                ) as cursor:
                    if user:
                        user.begin_progress(
                            _("Filter"), _("Applying ..."), self.get_number(db)
                        )

                    for handle, obj in cursor:
                        if user:
                            user.step_progress()

                        if handle in handles_out:
                            continue

                        if apply_logical_op(db, obj, self.flist) != self.invert:
                            final_list.append(handle)

        else:
            id_list = list(id_list)
            if user:
                user.begin_progress(_("Filter"), _("Applying ..."), len(id_list))
            for handle_data in id_list:
                if user:
                    user.step_progress()

                if tupleind is None:
                    handle = handle_data
                else:
                    handle = handle_data[tupleind]

                if handles_in is not None:
                    if handle not in handles_in:
                        continue
                elif handle in handles_out:
                    continue

                obj = self.get_object(db, handle)

                if apply_logical_op(db, obj, self.flist) != self.invert:
                    final_list.append(handle_data)

        if user:
            user.end_progress()

        return final_list

    def and_test(self, db, data: dict, flist):
        return all(rule.apply_to_one(db, data) for rule in flist)

    def one_test(self, db, data: dict, flist):
        found_one = False
        for rule in flist:
            if rule.apply_to_one(db, data):
                if found_one:
                    return False  # There can be only one!
                found_one = True
        return found_one

    def or_test(self, db, data: dict, flist):
        return any(rule.apply_to_one(db, data) for rule in flist)

    def get_logical_op(self):
        return self.logical_op

    def apply_to_one(self, db, data: dict) -> bool:
        """
        Filter-level apply rules to single data item.
        """
        if self.logical_op == "and":
            res = self.and_test(db, data, self.flist)
        elif self.logical_op == "or":
            res = self.or_test(db, data, self.flist)
        elif self.logical_op == "one":
            res = self.one_test(db, data, self.flist)
        else:
            raise Exception("invalid operator: %r" % self.logical_op)
        return res != self.invert

    def apply(self, db, id_list=None, tupleind=None, user=None, tree=False):
        """
        Apply the filter using db.
        If id_list given, the handles in id_list are used. If not given
        a database cursor will be used over all entries.

        If tupleind is given, id_list is supposed to consist of a list of
        tuples, with the handle being index tupleind. So
        handle_0 = id_list[0][tupleind]

        user is optional. If present it must be an instance of a User class.

        :Returns: if id_list given, it is returned with the items that
                do not match the filter, filtered out.
                if id_list not given, all items in the database that
                match the filter are returned as a list of handles
        """
        start_time = time.time()
        for rule in self.flist:
            rule.requestprepare(db, user)
        LOG.debug("Prepare time: %s seconds", time.time() - start_time)

        if self.logical_op == "and":
            apply_logical_op = self.and_test
        elif self.logical_op == "or":
            apply_logical_op = self.or_test
        elif self.logical_op == "one":
            apply_logical_op = self.one_test
        else:
            raise Exception("invalid operator: %r" % self.logical_op)

        start_time = time.time()
        res = self.apply_logical_op_to_all(
            db, id_list, apply_logical_op, user, tupleind, tree
        )
        LOG.debug("Apply time: %s seconds", time.time() - start_time)

        for rule in self.flist:
            rule.requestreset()

        return res

    def get_object(self, db, handle):
        return db.get_person_from_handle(handle)


class GenericFamilyFilter(GenericFilter):
    def __init__(self, source=None):
        GenericFilter.__init__(self, source)

    def get_cursor(self, db):
        return db.get_family_cursor()

    def make_obj(self):
        return Family()

    def find_from_handle(self, db, handle):
        return db.get_family_from_handle(handle)

    def get_number(self, db):
        return db.get_number_of_families()

    def get_object(self, db, handle):
        return db.get_family_from_handle(handle)


class GenericEventFilter(GenericFilter):
    def __init__(self, source=None):
        GenericFilter.__init__(self, source)

    def get_cursor(self, db):
        return db.get_event_cursor()

    def make_obj(self):
        return Event()

    def find_from_handle(self, db, handle):
        return db.get_event_from_handle(handle)

    def get_number(self, db):
        return db.get_number_of_events()

    def get_object(self, db, handle):
        return db.get_event_from_handle(handle)


class GenericSourceFilter(GenericFilter):
    def __init__(self, source=None):
        GenericFilter.__init__(self, source)

    def get_cursor(self, db):
        return db.get_source_cursor()

    def make_obj(self):
        return Source()

    def find_from_handle(self, db, handle):
        return db.get_source_from_handle(handle)

    def get_number(self, db):
        return db.get_number_of_sources()

    def get_object(self, db, handle):
        return db.get_source_from_handle(handle)


class GenericCitationFilter(GenericFilter):
    def __init__(self, source=None):
        GenericFilter.__init__(self, source)

    def get_cursor(self, db):
        return db.get_citation_cursor()

    def get_tree_cursor(self, db):
        return db.get_citation_cursor()

    def make_obj(self):
        return Citation()

    def find_from_handle(self, db, handle):
        return db.get_citation_from_handle(handle)

    def get_number(self, db):
        return db.get_number_of_citations()

    def get_object(self, db, handle):
        return db.get_citation_from_handle(handle)


class GenericPlaceFilter(GenericFilter):
    def __init__(self, source=None):
        GenericFilter.__init__(self, source)

    def get_cursor(self, db):
        return db.get_place_cursor()

    def get_tree_cursor(self, db):
        return db.get_place_tree_cursor()

    def make_obj(self):
        return Place()

    def find_from_handle(self, db, handle):
        return db.get_place_from_handle(handle)

    def get_number(self, db):
        return db.get_number_of_places()

    def get_object(self, db, handle):
        return db.get_place_from_handle(handle)


class GenericMediaFilter(GenericFilter):
    def __init__(self, source=None):
        GenericFilter.__init__(self, source)

    def get_cursor(self, db):
        return db.get_media_cursor()

    def make_obj(self):
        return Media()

    def find_from_handle(self, db, handle):
        return db.get_media_from_handle(handle)

    def get_number(self, db):
        return db.get_number_of_media()

    def get_object(self, db, handle):
        return db.get_media_from_handle(handle)


class GenericRepoFilter(GenericFilter):
    def __init__(self, source=None):
        GenericFilter.__init__(self, source)

    def get_cursor(self, db):
        return db.get_repository_cursor()

    def make_obj(self):
        return Repository()

    def find_from_handle(self, db, handle):
        return db.get_repository_from_handle(handle)

    def get_number(self, db):
        return db.get_number_of_repositories()

    def get_object(self, db, handle):
        return db.get_repository_from_handle(handle)


class GenericNoteFilter(GenericFilter):
    def __init__(self, source=None):
        GenericFilter.__init__(self, source)

    def get_cursor(self, db):
        return db.get_note_cursor()

    def make_obj(self):
        return Note()

    def find_from_handle(self, db, handle):
        return db.get_note_from_handle(handle)

    def get_number(self, db):
        return db.get_number_of_notes()

    def get_object(self, db, handle):
        return db.get_note_from_handle(handle)


def GenericFilterFactory(namespace):
    if namespace == "Person":
        return GenericFilter
    elif namespace == "Family":
        return GenericFamilyFilter
    elif namespace == "Event":
        return GenericEventFilter
    elif namespace == "Source":
        return GenericSourceFilter
    elif namespace == "Citation":
        return GenericCitationFilter
    elif namespace == "Place":
        return GenericPlaceFilter
    elif namespace == "Media":
        return GenericMediaFilter
    elif namespace == "Repository":
        return GenericRepoFilter
    elif namespace == "Note":
        return GenericNoteFilter


class DeferredFilter(GenericFilter):
    """
    Filter class allowing for deferred translation of the filter name
    """

    def __init__(self, filter_name, person_name):
        GenericFilter.__init__(self, None)
        self.name_pair = [filter_name, person_name]

    def get_name(self, ulocale=glocale):
        """
        return the filter name, possibly translated

        If ulocale is passed in (a :class:`.GrampsLocale`) then
        the translated value will be returned instead.

        :param ulocale: allow deferred translation of strings
        :type ulocale: a :class:`.GrampsLocale` instance
        """
        self._ = ulocale.translation.gettext
        if self.name_pair[1]:
            return self._(self.name_pair[0]) % self.name_pair[1]
        return self._(self.name_pair[0])


class DeferredFamilyFilter(GenericFamilyFilter):
    """
    Filter class allowing for deferred translation of the filter name
    """

    def __init__(self, filter_name, family_name):
        GenericFamilyFilter.__init__(self, None)
        self.name_pair = [filter_name, family_name]

    def get_name(self, ulocale=glocale):
        """
        return the filter name, possibly translated

        If ulocale is passed in (a :class:`.GrampsLocale`) then
        the translated value will be returned instead.

        :param ulocale: allow deferred translation of strings
        :type ulocale: a :class:`.GrampsLocale` instance
        """
        self._ = ulocale.translation.gettext
        if self.name_pair[1]:
            return self._(self.name_pair[0]) % self.name_pair[1]
        return self._(self.name_pair[0])



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from __future__ import annotations
from ....const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .. import Rule


# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from typing import Set
from ....lib import Person
from ....db import Database


# -------------------------------------------------------------------------
#
# IsDescendantOf
#
# -------------------------------------------------------------------------
class IsDescendantOf(Rule):
    """Rule that checks for a person that is a descendant
    of a specified person"""

    labels = [_("ID:"), _("Inclusive:")]
    name = _("Descendants of <person>")
    category = _("Descendant filters")
    description = _("Matches all descendants for the specified person")

    def prepare(self, db: Database, user):
        self.db = db
        self.selected_handles: Set[str] = set()
        try:
            first = False if int(self.list[1]) else True
        except IndexError:
            first = True
        try:
            root_person = db.get_person_from_gramps_id(self.list[0])
            self.init_list(root_person, first)
        except:
            pass

    def reset(self):
        self.selected_handles.clear()

    def apply_to_one(self, db: Database, person: Person) -> bool:
        return person.handle in self.selected_handles

    def init_list(self, person: Person | None, first: bool) -> None:
        if not person or person.handle in self.selected_handles:
            # if we have been here before, skip
            return
        if not first:
            self.selected_handles.add(person.handle)

        for fam_id in person.family_list:
            fam = self.db.get_family_from_handle(fam_id)
            if fam:
                for child_ref in fam.child_ref_list:
                    self.init_list(self.db.get_person_from_handle(child_ref.ref), False)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2003-2007 Donald N. Allingham
# Copyright (C) 2007-2012 Brian G. Matherly
# Copyright (C) 2010      Jakim Friant
# Copyright (C) 2012-2016 Paul Franklin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Timeline Chart
"""

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
from gramps.gen.plug.menu import PersonOption, FilterOption, EnumeratedListOption
from gramps.gen.plug.report import Report
from gramps.gen.plug.report import utils
from gramps.gen.plug.report import MenuReportOptions
from gramps.gen.plug.report import stdoptions
from gramps.gen.plug.docgen import (
    FontStyle,
    ParagraphStyle,
    GraphicsStyle,
    FONT_SANS_SERIF,
    DASHED,
    PARA_ALIGN_CENTER,
    IndexMark,
    INDEX_TYPE_TOC,
)
from gramps.gen.sort import Sort
from gramps.gen.config import config
from gramps.gen.utils.db import get_birth_or_fallback, get_death_or_fallback
from gramps.gen.proxy import CacheProxyDb
from gramps.gen.lib import Date

# ------------------------------------------------------------------------
#
# private constants
#
# ------------------------------------------------------------------------


# _T_ is a gramps-defined keyword -- see po/update_po.py and po/genpot.sh
def _T_(value, context=""):  # enable deferred translations
    return "%s\x04%s" % (context, value) if context else value


# ------------------------------------------------------------------------
#
# Private Functions
#
# ------------------------------------------------------------------------
def _get_sort_functions(sort):
    return [
        (_T_("Birth Date", "sorted by"), sort.by_birthdate_key),
        (_T_("Name", "sorted by"), sort.by_last_name_key),
    ]


# ------------------------------------------------------------------------
#
# TimeLine
#
# ------------------------------------------------------------------------
class TimeLine(Report):
    """TimeLine Report"""

    def __init__(self, database, options, user):
        """
        Create the Timeline object that produces the report.

        The arguments are:

        database        - the Gramps database instance
        options         - instance of the Options class for this report
        user            - instance of gen.user.User()

        This report needs the following parameters (class variables)
        that come in the options class.

        filter    - Filter to be applied to the people of the database.
                    The option class carries its number, and the function
                    returning the list of filters.
        sortby        - Sorting method to be used.
        name_format   - Preferred format to display names
        incl_private  - Whether to include private data
        living_people - How to handle living people
        years_past_death - Consider as living this many years after death
        """
        Report.__init__(self, database, options, user)
        self._user = user
        menu = options.menu

        self.set_locale(options.menu.get_option_by_name("trans").get_value())

        stdoptions.run_private_data_option(self, menu)
        living_opt = stdoptions.run_living_people_option(self, menu, self._locale)
        self.database = CacheProxyDb(self.database)

        self.filter = menu.get_option_by_name("filter").get_filter()
        self.fil_name = "(%s)" % self.filter.get_name(self._locale)

        living_value = menu.get_option_by_name("living_people").get_value()
        for value, description in living_opt.get_items(xml_items=True):
            if value == living_value:
                living_desc = self._(description)
                break
        self.living_desc = self._("(Living people: %(option_name)s)") % {
            "option_name": living_desc
        }

        stdoptions.run_name_format_option(self, menu)

        sort_func_num = menu.get_option_by_name("sortby").get_value()
        sort_functions = _get_sort_functions(Sort(self.database))
        self.sort_name = self._(sort_functions[sort_func_num][0])
        self.sort_func = sort_functions[sort_func_num][1]
        self.calendar = config.get("preferences.calendar-format-report")
        self.plist = []
        self.header = 2.6

    def write_report(self):
        # Apply the filter
        self.plist = self.filter.apply(
            self.database, self.database.iter_person_handles(), user=self._user
        )

        # Find the range of dates to include
        (low, high) = self.find_year_range()

        # Generate the actual timeline
        self.generate_timeline(low, high)

    def generate_timeline(self, low, high):
        """generate the timeline"""
        st_size = self.name_size()
        style_sheet = self.doc.get_style_sheet()
        font = style_sheet.get_paragraph_style("TLG-Name").get_font()
        incr = utils.pt2cm(font.get_size())
        pad = incr * 0.75
        _x1, _x2, _y1, _y2 = (0, 0, 0, 0)
        start = st_size + 0.5
        stop = self.doc.get_usable_width() - 0.5
        size = stop - start
        self.header = 2.6

        # Sort the people as requested
        with self._user.progress(_("Timeline"), _("Sorting dates..."), 0) as step:
            self.plist.sort(key=self.sort_func)

        self.doc.start_page()
        self.build_grid(low, high, start, stop, True)

        index = 1
        current = 1

        length = len(self.plist)

        with self._user.progress(
            _("Timeline"), _("Calculating timeline..."), length
        ) as step:
            for p_id in self.plist:
                person = self.database.get_person_from_handle(p_id)
                birth = get_birth_or_fallback(self.database, person)
                if birth:
                    bth = birth.get_date_object()
                    bth = bth.to_calendar(self.calendar).get_year()
                else:
                    bth = None

                death = get_death_or_fallback(self.database, person)
                if death:
                    dth = death.get_date_object()
                    dth = dth.to_calendar(self.calendar).get_year()
                else:
                    dth = None

                dname = self._name_display.display(person)
                mark = utils.get_person_mark(self.database, person)
                self.doc.draw_text(
                    "TLG-text",
                    dname,
                    incr + pad,
                    self.header + (incr + pad) * index,
                    mark,
                )

                _y1 = self.header + (pad + incr) * index
                _y2 = self.header + ((pad + incr) * index) + incr
                _y3 = (_y1 + _y2) / 2.0
                w05 = 0.05

                if bth:
                    start_offset = (float(bth - low) / float(high - low)) * size
                    _x1 = start + start_offset
                    path = [(_x1, _y1), (_x1 + w05, _y3), (_x1, _y2), (_x1 - w05, _y3)]
                    self.doc.draw_path("TLG-line", path)

                if dth:
                    start_offset = (float(dth - low) / float(high - low)) * size
                    _x1 = start + start_offset
                    path = [(_x1, _y1), (_x1 + w05, _y3), (_x1, _y2), (_x1 - w05, _y3)]
                    self.doc.draw_path("TLG-solid", path)

                if bth and dth:
                    start_offset = ((float(bth - low) / float(high - low)) * size) + w05
                    stop_offset = ((float(dth - low) / float(high - low)) * size) - w05

                    _x1 = start + start_offset
                    _x2 = start + stop_offset
                    self.doc.draw_line("open", _x1, _y3, _x2, _y3)

                if (_y2 + incr) >= self.doc.get_usable_height():
                    if current != length:
                        self.doc.end_page()
                        self.doc.start_page()
                        self.build_grid(low, high, start, stop)
                    index = 1
                    _x1, _x2, _y1, _y2 = (0, 0, 0, 0)
                else:
                    index += 1
                current += 1
                step()
            self.doc.end_page()

    def build_grid(self, year_low, year_high, start_pos, stop_pos, toc=False):
        """
        Draws the grid outline for the chart. Sets the document label,
        draws the vertical lines, and adds the year labels. Arguments
        are:

        year_low  - lowest year on the chart
        year_high - highest year on the chart
        start_pos - x position of the lowest leftmost grid line
        stop_pos  - x position of the rightmost grid line
        """
        self.draw_title(toc)
        self.draw_columns(start_pos, stop_pos)
        if year_high is not None and year_low is not None:
            self.draw_year_headings(year_low, year_high, start_pos, stop_pos)
        else:
            self.draw_no_date_heading()

    def draw_columns(self, start_pos, stop_pos):
        """
        Draws the columns out of vertical lines.

        start_pos - x position of the lowest leftmost grid line
        stop_pos  - x position of the rightmost grid line
        """
        top_y = self.header
        bottom_y = self.doc.get_usable_height()
        delta = (stop_pos - start_pos) / 5
        for val in range(0, 6):
            xpos = start_pos + (val * delta)
            self.doc.draw_line("TLG-grid", xpos, top_y, xpos, bottom_y)

    def draw_title(self, toc):
        """
        Draws the title for the page.
        """
        width = self.doc.get_usable_width()
        title = "%(str1)s -- %(str2)s" % {
            "str1": self._("Timeline Chart"),
            # feature request 2356: avoid genitive form
            "str2": self._("Sorted by %s") % self.sort_name,
        }
        title3 = self.living_desc
        mark = None
        if toc:
            mark = IndexMark(title, INDEX_TYPE_TOC, 1)
        self.doc.center_text("TLG-title", title, width / 2.0, 0, mark)
        style_sheet = self.doc.get_style_sheet()
        title_font = style_sheet.get_paragraph_style("TLG-Title").get_font()
        title_y = 1.2 - (utils.pt2cm(title_font.get_size()) * 1.2)
        self.doc.center_text("TLG-title", self.fil_name, width / 2.0, title_y)
        title_y = 1.8 - (utils.pt2cm(title_font.get_size()) * 1.2)
        self.doc.center_text("TLG-title", title3, width / 2.0, title_y)

    def draw_year_headings(self, year_low, year_high, start_pos, stop_pos):
        """
        Draws the column headings (years) for the page.
        """
        style_sheet = self.doc.get_style_sheet()
        label_font = style_sheet.get_paragraph_style("TLG-Label").get_font()
        label_y = self.header - (utils.pt2cm(label_font.get_size()) * 1.2)
        incr = (year_high - year_low) / 5
        delta = (stop_pos - start_pos) / 5
        for val in range(0, 6):
            xpos = start_pos + (val * delta)
            year_str = self._get_date(Date(year_low + int(incr * val)))
            self.doc.center_text("TLG-label", year_str, xpos, label_y)

    def draw_no_date_heading(self):
        """
        Draws a single heading that says "No Date Information"
        """
        width = self.doc.get_usable_width()
        style_sheet = self.doc.get_style_sheet()
        label_font = style_sheet.get_paragraph_style("TLG-Label").get_font()
        label_y = self.header - (utils.pt2cm(label_font.get_size()) * 1.2)
        self.doc.center_text(
            "TLG-label", self._("No Date Information"), width / 2.0, label_y
        )

    def find_year_range(self):
        """
        Finds the range of years that will be displayed on the chart.

        Returns a tuple of low and high years. If no dates are found, the
        function returns (None, None).
        """
        low = None
        high = None

        def min_max_year(low, high, year):
            """convenience function"""
            if year is not None and year != 0:
                if low is not None:
                    low = min(low, year)
                else:
                    low = year
                if high is not None:
                    high = max(high, year)
                else:
                    high = year
            return (low, high)

        with self._user.progress(
            _("Timeline"), _("Finding date range..."), len(self.plist)
        ) as step:
            for p_id in self.plist:
                person = self.database.get_person_from_handle(p_id)
                birth = get_birth_or_fallback(self.database, person)
                if birth:
                    bth = birth.get_date_object()
                    bth = bth.to_calendar(self.calendar).get_year()
                    (low, high) = min_max_year(low, high, bth)

                death = get_death_or_fallback(self.database, person)
                if death:
                    dth = death.get_date_object()
                    dth = dth.to_calendar(self.calendar).get_year()
                    (low, high) = min_max_year(low, high, dth)
                step()

            # round the dates to the nearest decade
            if low is not None:
                low = int((low / 10)) * 10
            else:
                low = high

            if high is not None:
                high = int(((high + 9) / 10)) * 10
            else:
                high = low

            # Make sure the difference is a multiple of 50 so
            # all year ranges land on a decade.
            if low is not None and high is not None:
                low -= 50 - ((high - low) % 50)

        return (low, high)

    def name_size(self):
        """get the length of the name"""
        self.plist = self.filter.apply(
            self.database, self.database.iter_person_handles(), user=self._user
        )

        style_sheet = self.doc.get_style_sheet()
        gstyle = style_sheet.get_draw_style("TLG-text")
        pname = gstyle.get_paragraph_style()
        pstyle = style_sheet.get_paragraph_style(pname)
        font = pstyle.get_font()

        size = 0
        for p_id in self.plist:
            person = self.database.get_person_from_handle(p_id)
            dname = self._name_display.display(person)
            size = max(self.doc.string_width(font, dname), size)
        return utils.pt2cm(size)


# ------------------------------------------------------------------------
#
# TimeLineOptions
#
# ------------------------------------------------------------------------
class TimeLineOptions(MenuReportOptions):
    """Options for the TimeLine Report"""

    def __init__(self, name, dbase):
        self.__pid = None
        self.__filter = None
        self.__db = dbase
        self._nf = None
        MenuReportOptions.__init__(self, name, dbase)

    def get_subject(self):
        """Return a string that describes the subject of the report."""
        return self.__filter.get_filter().get_name()

    def add_menu_options(self, menu):
        category_name = _("Report Options")

        self.__filter = FilterOption(_("Filter"), 0)
        self.__filter.set_help(_("Determines what people are included in the report"))
        menu.add_option(category_name, "filter", self.__filter)
        self.__filter.connect("value-changed", self.__filter_changed)

        self.__pid = PersonOption(_("Filter Person"))
        self.__pid.set_help(_("The center person for the filter"))
        menu.add_option(category_name, "pid", self.__pid)
        self.__pid.connect("value-changed", self.__update_filters)

        sortby = EnumeratedListOption(_("Sort by"), 0)
        idx = 0
        for item in _get_sort_functions(Sort(self.__db)):
            sortby.add_item(idx, _(item[0]))
            idx += 1
        sortby.set_help(_("Sorting method to use"))
        menu.add_option(category_name, "sortby", sortby)

        category_name = _("Report Options (2)")

        self._nf = stdoptions.add_name_format_option(menu, category_name)
        self._nf.connect("value-changed", self.__update_filters)

        self.__update_filters()

        stdoptions.add_private_data_option(menu, category_name)

        stdoptions.add_living_people_option(menu, category_name)

        stdoptions.add_localization_option(menu, category_name)

    def __update_filters(self):
        """
        Update the filter list based on the selected person
        """
        gid = self.__pid.get_value()
        person = self.__db.get_person_from_gramps_id(gid)
        nfv = self._nf.get_value()
        filter_list = utils.get_person_filters(
            person, include_single=False, name_format=nfv
        )
        self.__filter.set_filters(filter_list)

    def __filter_changed(self):
        """
        Handle filter change. If the filter is not specific to a person,
        disable the person option
        """
        filter_value = self.__filter.get_value()
        if filter_value == 0:  # "Entire Database" (as "include_single=False")
            self.__pid.set_available(False)
        else:
            # The other filters need a center person (assume custom ones too)
            self.__pid.set_available(True)

    def make_default_style(self, default_style):
        """Make the default output style for the Timeline report."""
        # Paragraph Styles
        fstyle = FontStyle()
        fstyle.set_size(10)
        fstyle.set_type_face(FONT_SANS_SERIF)
        pstyle = ParagraphStyle()
        pstyle.set_font(fstyle)
        pstyle.set_description(_("The basic style used for the text display."))
        default_style.add_paragraph_style("TLG-Name", pstyle)

        fstyle = FontStyle()
        fstyle.set_size(8)
        fstyle.set_type_face(FONT_SANS_SERIF)
        pstyle = ParagraphStyle()
        pstyle.set_font(fstyle)
        pstyle.set_alignment(PARA_ALIGN_CENTER)
        pstyle.set_description(_("The style used for the section headers."))
        default_style.add_paragraph_style("TLG-Label", pstyle)

        fstyle = FontStyle()
        fstyle.set_size(14)
        fstyle.set_type_face(FONT_SANS_SERIF)
        pstyle = ParagraphStyle()
        pstyle.set_font(fstyle)
        pstyle.set_alignment(PARA_ALIGN_CENTER)
        pstyle.set_description(_("The style used for the title."))
        default_style.add_paragraph_style("TLG-Title", pstyle)

        """
        Graphic Styles
            TLG-grid  - 0.5pt wide line dashed line. Used for the lines that
                        make up the grid.
            TLG-line  - 0.5pt wide line. Used for the line connecting two
                        endpoints and for the birth marker.
            TLG-solid - 0.5pt line with a black fill color. Used for the date of
                        death marker.
            TLG-text  - Contains the TLG-Name paragraph style used for the
                        individual's name.
            TLG-title - Contains the TLG-Title paragraph style used for the
                        title of the document.
            TLG-label - Contains the TLG-Label paragraph style used for the year
                        label's in the document.
        """
        gstyle = GraphicsStyle()
        gstyle.set_line_width(0.5)
        gstyle.set_color((0, 0, 0))
        default_style.add_draw_style("TLG-line", gstyle)

        gstyle = GraphicsStyle()
        gstyle.set_line_width(0.5)
        gstyle.set_color((0, 0, 0))
        gstyle.set_fill_color((0, 0, 0))
        default_style.add_draw_style("TLG-solid", gstyle)

        gstyle = GraphicsStyle()
        gstyle.set_line_width(0.5)
        gstyle.set_color((0, 0, 0))
        gstyle.set_fill_color((255, 255, 255))
        default_style.add_draw_style("open", gstyle)

        gstyle = GraphicsStyle()
        gstyle.set_line_width(0.5)
        gstyle.set_line_style(DASHED)
        gstyle.set_color((0, 0, 0))
        default_style.add_draw_style("TLG-grid", gstyle)

        gstyle = GraphicsStyle()
        gstyle.set_paragraph_style("TLG-Name")
        gstyle.set_color((0, 0, 0))
        gstyle.set_fill_color((255, 255, 255))
        gstyle.set_line_width(0)
        default_style.add_draw_style("TLG-text", gstyle)

        gstyle = GraphicsStyle()
        gstyle.set_paragraph_style("TLG-Title")
        gstyle.set_color((0, 0, 0))
        gstyle.set_fill_color((255, 255, 255))
        gstyle.set_line_width(0)
        default_style.add_draw_style("TLG-title", gstyle)

        gstyle = GraphicsStyle()
        gstyle.set_paragraph_style("TLG-Label")
        gstyle.set_color((0, 0, 0))
        gstyle.set_fill_color((255, 255, 255))
        gstyle.set_line_width(0)
        default_style.add_draw_style("TLG-label", gstyle)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2010  Douglas Blank <doug.blank@gmail.com>
# Copyright (C) 2011 Rob G. Healey <robhealey1@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# ------------------------------------------------
#    python modules
# ------------------------------------------------
import os
import re
from gramps.gen.const import USER_DATA_VERSION, IMAGE_DIR, DATA_DIR, USER_CSS
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext


def make_css_dict(tup):
    """
    Basically, make a named tuple.
    """
    return {
        "id": tup[0],
        "user": tup[1],
        "translation": tup[2],
        "filename": tup[3],
        "navigation": tup[4],
        "images": tup[5],
        "javascript": tup[6],
    }


def load_on_reg(dbstate, uistate, plugin):
    """
    Runs when plugin is registered.
    """
    from functools import partial

    path_css = partial(os.path.join, DATA_DIR, "css")
    path_img = partial(os.path.join, IMAGE_DIR, "webstuff")
    path_img_48x48 = partial(os.path.join, IMAGE_DIR, "hicolor/48x48/actions")
    CSS_FILES = [
        # id, user selectable?, translated_name, option name, fullpath,
        # navigation target name, images, javascript
        # "default" is used as default
        # default style sheet in the options
        # Translators: Basic Ash style sheet
        ["Basic-Ash", 1, _("Basic-Ash"), path_css("Web_Basic-Ash.css"), None, [], []],
        # Translators: Basic Blue style sheet with navigation menus
        [
            "Basic-Blue",
            1,
            _("Basic-Blue"),
            path_css("Web_Basic-Blue.css"),
            None,
            [],
            [],
        ],
        # Translators: Basic Cypress style sheet
        [
            "Basic-Cypress",
            1,
            _("Basic-Cypress"),
            path_css("Web_Basic-Cypress.css"),
            None,
            [],
            [],
        ],
        # Translators: Basic Lilac style sheet
        [
            "Basic-Lilac",
            1,
            _("Basic-Lilac"),
            path_css("Web_Basic-Lilac.css"),
            None,
            [],
            [],
        ],
        # Translators: Basic Peach style sheet
        [
            "Basic-Peach",
            1,
            _("Basic-Peach"),
            path_css("Web_Basic-Peach.css"),
            None,
            [],
            [],
        ],
        # Translators: Basic Spruce style sheet
        [
            "Basic-Spruce",
            1,
            _("Basic-Spruce"),
            path_css("Web_Basic-Spruce.css"),
            None,
            [],
            [],
        ],
        # Translators: Mainz style sheet with its images
        [
            "Mainz",
            1,
            _("Mainz"),
            path_css("Web_Mainz.css"),
            None,
            [
                path_img("Web_Mainz_Bkgd.png"),
                path_img("Web_Mainz_Header.png"),
                path_img("Web_Mainz_Mid.png"),
                path_img("Web_Mainz_MidLight.png"),
            ],
            [],
        ],
        # Translators: Nebraska style sheet
        ["Nebraska", 1, _("Nebraska"), path_css("Web_Nebraska.css"), None, [], []],
        # Translators: Visually Impaired style sheet with its navigation menus
        [
            "Visually Impaired",
            1,
            _("Visually Impaired"),
            path_css("Web_Visually.css"),
            "narrative-menus.css",
            [],
            [],
        ],
        # ancestor tree style sheet and its images
        [
            "ancestortree",
            0,
            "ancestortree",
            path_css("ancestortree.css"),
            None,
            [path_img("Web_Gender_Female.png"), path_img("Web_Gender_Male.png")],
            [],
        ],
        # media reference regions style sheet
        ["behaviour", 0, "Behaviour", path_css("behaviour.css"), None, [], []],
        # media lightbox style sheet
        ["lightbox", 0, "", path_css("lightbox.css"), None, [], []],
        # media lightbox javascript
        ["lightbox_js", 0, "", path_css("lightbox.js"), None, [], []],
        # NarrativeMap stylesheet/ image for NarrativeWeb place maps
        ["NarrativeMaps", 0, "", path_css("narrative-maps.css"), None, [], []],
        # default printer style sheet
        [
            "Print-Default",
            0,
            "Print-Default",
            path_css("Web_Print-Default.css"),
            None,
            [],
            [],
        ],
        # Horizontal Navigation Menus Style Sheet
        [
            "Horizontal-Menus",
            0,
            "Horizontal Menus",
            path_css("Web_Horizontal-Menus.css"),
            None,
            [],
            [],
        ],
        # Vertical Navigation Menus Style Sheet
        [
            "Vertical-Menus",
            0,
            "Vertical Menus",
            path_css("Web_Vertical-Menus.css"),
            None,
            [],
            [],
        ],
        # WebKit/ Html5/ CSS3 Fade Navigation Menus Style Sheet
        [
            "Fade-Menus",
            0,
            "Fade In/ Out Menus",
            path_css("Web_Fade-Menus.css"),
            None,
            [],
            [],
        ],
        # WebKit/ Html5/ CSS3 Animated Drop Down Style Sheet
        [
            "Animated DropDown",
            0,
            "Animated DropDown",
            path_css("Web_Citations-Animated.css"),
            None,
            [],
            "https://ajax.googleapis.com/ajax/libs/jquery/1.7.1/jquery.min.js",
        ],
        # Source Page Citations Referents Outline Style sheet...
        [
            "Outline",
            0,
            "Outline Citations",
            path_css("Web_Citations-Outline.css"),
            None,
            [],
            [],
        ],
        # WebKit/ Html5/ CSS3 Drop Down Navigation Menus Style Sheet
        [
            "DropDown-Menus",
            0,
            "Drop Down Menus",
            path_css("Web_DropDown-Menus.css"),
            None,
            [],
            [],
        ],
        # no style sheet option
        ["No style sheet", 1, _("No style sheet"), [], None, [], []],
        # Document image
        [
            "Document",
            0,
            "Document",
            os.path.join(IMAGE_DIR, "document.png"),
            None,
            [],
            [],
        ],
        # blank
        ["Blank", 0, "Blank", path_img("blank.gif"), None, [], []],
        # all other images for use in NarrativeWeb
        [
            "All Images",
            0,
            "All Images",
            None,
            None,
            [path_img("blank.gif"), os.path.join(IMAGE_DIR, "document.png")],
            [],
        ],
        # Gramps Fav icon #2
        ["favicon2", 0, "FavIcon2", path_img("favicon2.ico"), None, [], []],
        # copyright image
        ["Copyright", 0, "Copyright", path_img("somerights20.gif"), None, [], []],
        # marker icon for openstreetmap
        ["marker", 0, "marker", path_img_48x48("gramps-geo-mainmap.png"), None, [], []],
    ]
    # If we add css user files, we must restart gramps to use them.
    if os.path.exists(USER_CSS):
        list_files = os.listdir(USER_CSS)
        for cssfile in list_files:
            if cssfile.endswith(".css"):
                css_f = cssfile.replace(".css", "")
                CSS_FILES.append(
                    [
                        "UsEr_" + css_f,
                        1,
                        css_f,
                        os.path.join(USER_CSS, cssfile),
                        None,
                        looking_for_urls_in_user_css(cssfile),
                        [],
                    ]
                )
    return CSS_FILES


def looking_for_urls_in_user_css(css_file):
    """
    At each time we find the tag url, we get the content and add it
    to the images list. This content must be local.
    """
    images = []
    cssfile = os.path.join(USER_CSS, css_file)
    with open(cssfile) as css:
        data = css.readlines()
        for line in data:
            if "url" in line:
                url = re.match(r".*url\((.*)\)", line)
                if url.group(1)[0:3] != "http":
                    img = url.group(1).replace("../images/", "")
                    img = os.path.join(USER_CSS, img)
                    if img not in images:
                        images.append("%s" % img)
    return images


def process_list(data):
    """
    Gather all of the web resources together, and allow override files
    if available.
    """
    retval = []
    for row in data:
        file = row[3]
        if file:
            dummy_path, filename = os.path.split(file)
            # is there a override file in the DATA_VERSION_DIR/webstuff?
            # eg, ~/.gramps/gramps34/webstuff/Web_Nebraska.css
            # if so, replace this one:
            override = os.path.join(USER_DATA_VERSION, "webstuff", filename)
            if os.path.exists(override):
                row[3] = override
        retval.append(row)
    # {"Mainz": {"id": "Mainz", "user":1, ...}}
    retdict = {}
    for css in retval:
        if css[0] in retdict:
            retdict[css[0]]["images"].append(css[5])
        else:
            retdict[css[0]] = make_css_dict(css)
    return retdict



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2011,2014  Nick Hall
# Copyright (C) 2011       Rob G. Healey <robhealey1@gmail.com>
# Copyright (C) 2022       Bruce Jackson
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import os
import logging

_LOG = logging.getLogger(".libmetadata")


# -------------------------------------------------------------------------
#
# GNOME modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk
import gi

gi.require_version("GExiv2", "0.10")
from gi.repository import GExiv2
from gi.repository import Gdk
from gi.repository import GdkPixbuf
from gi.repository import GObject


# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------

from gramps.gui.listmodel import ListModel, NOSORT, IMAGE as COL_IMAGE
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from gramps.gen.utils.place import conv_lat_lon
from fractions import Fraction
from gramps.gen.lib import Date
from gramps.gen.datehandler import displayer
from datetime import datetime

THUMBNAIL_IMAGE_SIZE = (50, 50)


def format_datetime(datestring):
    """
    Convert an exif timestamp into a string for display, using the
    standard Gramps date format.  Function not used for XMP Date Metatags:
    https://www.iptc.org/std/photometadata/specification/IPTC-PhotoMetadata#date-value-type
    """
    try:
        timestamp = datetime.strptime(datestring, "%Y:%m:%d %H:%M:%S")
    except ValueError:
        return _("Invalid format")

    date_part = Date()
    date_part.set_yr_mon_day(timestamp.year, timestamp.month, timestamp.day)
    date_str = displayer.display(date_part)
    time_str = _("%(hr)02d:%(min)02d:%(sec)02d") % {
        "hr": timestamp.hour,
        "min": timestamp.minute,
        "sec": timestamp.second,
    }
    return _("%(date)s %(time)s") % {"date": date_str, "time": time_str}


def format_gps(raw_dms, nsew):
    """
    Convert raw degrees, minutes, seconds and a direction
    reference into a string for display.
    """
    value = 0.0
    divisor = 1.0
    for val in raw_dms.split(" "):
        try:
            num = float(val.split("/")[0]) / float(val.split("/")[1])
        except (ValueError, IndexError):
            value = None
            break
        value += num / divisor
        divisor *= 60

    if nsew == "N":
        result = conv_lat_lon(str(value), "0", "DEG")[0]
    elif nsew == "S":
        result = conv_lat_lon("-" + str(value), "0", "DEG")[0]
    elif nsew == "E":
        result = conv_lat_lon("0", str(value), "DEG")[1]
    elif nsew == "W":
        result = conv_lat_lon("0", "-" + str(value), "DEG")[1]
    else:
        result = None

    return result if result is not None else _("Invalid format")


DESCRIPTION = _("Descriptive Tags")
DATE = _("Date and Time Tags")
PEOPLE = _("People Tags")
EVENT = _("Event Tags")
IMAGE = _("Image Tags")
CAMERA = _("Camera Information")
LOCATION = _("Location Tags")
ADVANCED = _("Advanced Tags")
RIGHTS = _("Rights Tags")
TAGGING = _("Keyword Tags")

"""
List of tags available to plugin can be found at the Exiv2 project
https://www.exiv2.org/metadata.html
"""

TAGS = [
    (DESCRIPTION, "Exif.Image.ImageDescription", None, None),
    (DESCRIPTION, "Exif.Image.XPSubject", None, None),
    (DESCRIPTION, "Exif.Image.XPComment", None, None),
    (DESCRIPTION, "Exif.Image.Rating", None, None),
    (DESCRIPTION, "Xmp.dc.title", None, None),
    (DESCRIPTION, "Xmp.dc.description", None, None),
    (DESCRIPTION, "Xmp.dc.subject", None, None),
    (DESCRIPTION, "Xmp.acdsee.caption", None, None),
    (DESCRIPTION, "Xmp.acdsee.notes", None, None),
    (DESCRIPTION, "Iptc.Application2.Caption", None, None),
    (DESCRIPTION, "Exif.Photo.UserComment", None, None),
    (DESCRIPTION, "Xmp.iptcExt.AOTitle", None, None),
    (DATE, "Exif.Photo.DateTimeOriginal", None, format_datetime),
    (DATE, "Exif.Photo.DateTimeDigitized", None, format_datetime),
    (DATE, "Exif.Image.DateTime", None, format_datetime),
    (DATE, "Exif.Image.TimeZoneOffset", None, None),
    (DATE, "Xmp.Xmp.CreateDate", None, None),
    (DATE, "Xmp.photoshop.DateCreated", None, None),
    (PEOPLE, "Xmp.mwg-rs.Regions/mwg-rs:RegionList[1]/mwg-rs:Name", None, None),
    (PEOPLE, "Xmp.mwg-rs.Regions", None, None),
    (PEOPLE, "Xmp.iptcExt.PersonInImage", None, None),
    (EVENT, "Xmp.iptcExt.Event", None, None),
    (LOCATION, "Xmp.iptcExt.LocationShown", None, None),
    (LOCATION, "Exif.GPSInfo.GPSLatitude", "Exif.GPSInfo.GPSLatitudeRef", format_gps),
    (LOCATION, "Exif.GPSInfo.GPSLongitude", "Exif.GPSInfo.GPSLongitudeRef", format_gps),
    (LOCATION, "Exif.GPSInfo.GPSAltitude", "Exif.GPSInfo.GPSAltitudeRef", None),
    (LOCATION, "Exif.GPSInfo.GPSTimeStamp", None, None),
    (LOCATION, "Exif.GPSInfo.GPSSatellites", None, None),
    (TAGGING, "Exif.Image.XPKeywords", None, None),
    (TAGGING, "Iptc.Application2.Keywords", None, None),
    (TAGGING, "Xmp.mwg-kw.Hierarchy", None, None),
    (TAGGING, "Xmp.mwg-kw.Keywords", None, None),
    (TAGGING, "Xmp.digiKam.TagsList", None, None),
    (TAGGING, "Xmp.MicrosoftPhoto.LastKeywordXMP", None, None),
    (TAGGING, "Xmp.MicrosoftPhoto.LastKeywordIPTC", None, None),
    (TAGGING, "Xmp.lr.hierarchicalSubject", None, None),
    (TAGGING, "Xmp.acdsee.categories", None, None),
    (IMAGE, "Exif.Image.DocumentName", None, None),
    (IMAGE, "Exif.Photo.PixelXDimension", None, None),
    (IMAGE, "Exif.Photo.PixelYDimension", None, None),
    (IMAGE, "Exif.Image.XResolution", "Exif.Image.ResolutionUnit", None),
    (IMAGE, "Exif.Image.YResolution", "Exif.Image.ResolutionUnit", None),
    (IMAGE, "Exif.Image.Orientation", None, None),
    (IMAGE, "Exif.Photo.ColorSpace", None, None),
    (IMAGE, "Exif.Image.YCbCrPositioning", None, None),
    (IMAGE, "Exif.Photo.ComponentsConfiguration", None, None),
    (IMAGE, "Exif.Image.Compression", None, None),
    (IMAGE, "Exif.Photo.CompressedBitsPerPixel", None, None),
    (IMAGE, "Exif.Image.PhotometricInterpretation", None, None),
    (RIGHTS, "Exif.Image.Copyright", None, None),
    (RIGHTS, "Exif.Image.Artist", None, None),
    (RIGHTS, "Xmp.xmpRights.Owner", None, None),
    (RIGHTS, "Xmp.xmpRights.UsageTerms", None, None),
    (RIGHTS, "Xmp.xmpRights.WebStatement", None, None),
    (CAMERA, "Exif.Image.Make", None, None),
    (CAMERA, "Exif.Image.Model", None, None),
    (CAMERA, "Exif.Photo.FNumber", None, None),
    (CAMERA, "Exif.Photo.ExposureTime", None, None),
    (CAMERA, "Exif.Photo.ISOSpeedRatings", None, None),
    (CAMERA, "Exif.Photo.FocalLength", None, None),
    (CAMERA, "Exif.Photo.FocalLengthIn35mmFilm", None, None),
    (CAMERA, "Exif.Photo.MaxApertureValue", None, None),
    (CAMERA, "Exif.Photo.MeteringMode", None, None),
    (CAMERA, "Exif.Photo.ExposureProgram", None, None),
    (CAMERA, "Exif.Photo.ExposureBiasValue", None, None),
    (CAMERA, "Exif.Photo.Flash", None, None),
    (CAMERA, "Exif.Image.FlashEnergy", None, None),
    (CAMERA, "Exif.Image.SelfTimerMode", None, None),
    (CAMERA, "Exif.Image.SubjectDistance", None, None),
    (CAMERA, "Exif.Photo.Contrast", None, None),
    (CAMERA, "Exif.Photo.LightSource", None, None),
    (CAMERA, "Exif.Photo.Saturation", None, None),
    (CAMERA, "Exif.Photo.Sharpness", None, None),
    (CAMERA, "Exif.Photo.WhiteBalance", None, None),
    (CAMERA, "Exif.Photo.DigitalZoomRatio", None, None),
    (ADVANCED, "Exif.Image.Software", None, None),
    (ADVANCED, "Exif.Photo.ImageUniqueID", None, None),
    (ADVANCED, "Exif.Image.CameraSerialNumber", None, None),
    (ADVANCED, "Exif.Photo.ExifVersion", None, None),
    (ADVANCED, "Exif.Photo.FlashpixVersion", None, None),
    (ADVANCED, "Exif.Image.ExifTag", None, None),
    (ADVANCED, "Exif.Image.GPSTag", None, None),
    (ADVANCED, "Exif.Image.BatteryLevel", None, None),
]


class MetadataView(Gtk.TreeView):
    def __init__(self):
        Gtk.TreeView.__init__(self)
        self.sections = {}
        titles = [
            (_("Namespace"), 0, 150),
            (_("Label"), 1, 150),
            (" ", NOSORT, 60, COL_IMAGE),
            (_("Value"), NOSORT, 325),
        ]

        self.model = ListModel(self, titles, list_mode="tree")

    def display_exif_tags(self, image_path):
        """
        Display the exif tags.
        """
        self.sections = {}

        # set fixed_height_mode to FALSE so thumbnails are not truncated.
        self.model.tree.set_fixed_height_mode(False)
        self.model.clear()

        if not os.path.exists(image_path):
            return False

        retval = False
        with open(image_path, "rb") as fd:
            try:
                buf = fd.read()
                metadata = GExiv2.Metadata()
                metadata.open_buf(buf)
                self.pixbuf = GdkPixbuf.Pixbuf.new_from_file(image_path)
                get_human = metadata.get_tag_interpreted_string

                for section, key, key2, func in TAGS:
                    if not key in self.__get_all_tags(metadata):
                        continue

                    if func is not None:
                        if key2 is None:
                            human_value = func(metadata[key])
                        else:
                            if key2 in self.__get_all_tags(metadata):
                                human_value = func(metadata[key], metadata[key2])
                            else:
                                human_value = func(metadata[key], None)
                    else:
                        human_value = get_human(key)
                        if key2 in self.__get_all_tags(metadata):
                            human_value += " " + get_human(key2)

                    if human_value is None:
                        human_value = ""

                    # If first named region is found - find all named regions
                    if key == "Xmp.mwg-rs.Regions/mwg-rs:RegionList[1]/mwg-rs:Name":
                        self.__get_named_regions(metadata)
                        continue

                    label = metadata.get_tag_label(key)
                    namespace = self.__get_tag_namespace(key)

                    node = self.__add_section(section)
                    self.model.add([namespace, label, None, human_value], node=node)

                    self.model.tree.expand_all()
                    retval = self.model.count > 0
            except:
                pass

        return retval

    def __add_section(self, section):
        """
        Add the section heading node to the model.
        """
        if section not in self.sections:
            node = self.model.add([section, "", None, ""])
            self.sections[section] = node
        else:
            node = self.sections[section]
        return node

    def get_has_data(self, image_path):
        """
        Return True if the gramplet has data, else return False.
        """
        if not os.path.exists(image_path):
            return False
        with open(image_path, "rb") as fd:
            retval = False
            try:
                buf = fd.read()
                metadata = GExiv2.Metadata()
                metadata.open_buf(buf)
                for tag in TAGS:
                    if tag in self.__get_all_tags(metadata):
                        retval = True
                        break
            except:
                pass

        return retval

    def __get_all_tags(self, metadata):
        """
        Return a list of all XMP, IPTC and EXIF tags in the media file
        """
        tag_list = (
            metadata.get_exif_tags()
            + metadata.get_xmp_tags()
            + metadata.get_iptc_tags()
        )

        return tag_list

    def __get_named_regions(self, metadata):
        """
        Retrieve all XMP named regions in an image and populate the treeview row.
        """

        region_tag = "Xmp.mwg-rs.Regions/mwg-rs:RegionList[%s]/"
        region_name = region_tag + "mwg-rs:Name"
        region_x = region_tag + "mwg-rs:Area/stArea:x"
        region_y = region_tag + "mwg-rs:Area/stArea:y"
        region_w = region_tag + "mwg-rs:Area/stArea:w"
        region_h = region_tag + "mwg-rs:Area/stArea:h"

        pixbuf_width = self.pixbuf.get_width()
        pixbuf_height = self.pixbuf.get_height()

        i = 1
        while True:
            name = metadata.get(region_name % i)
            region_name_display = region_name % i

            if name is None:
                break
            try:
                x = float(metadata.get(region_x % i)) * pixbuf_width
                y = float(metadata.get(region_y % i)) * pixbuf_height
                w = float(metadata.get(region_w % i)) * pixbuf_width
                h = float(metadata.get(region_h % i)) * pixbuf_height
            except ValueError:
                x = pixbuf_width / 2
                y = pixbuf_height / 2
                w = pixbuf_width
                h = pixbuf_height

            # ensure region does not exceed bounds of image
            region_p1 = x - (w / 2)
            if region_p1 < 0:
                region_p1 = 0
            region_p2 = y - (h / 2)
            if region_p2 < 0:
                region_p2 = 0
            region_p3 = x + (w / 2)
            if region_p3 > pixbuf_width:
                region_p3 = pixbuf_width
            region_p4 = y + (h / 2)
            if region_p4 > pixbuf_height:
                region_p4 = pixbuf_height

            region = (region_p1, region_p2, region_p3, region_p4)
            person_thumbnail = self.__get_thumbnail(region, THUMBNAIL_IMAGE_SIZE)

            label = metadata.get_tag_label(region_name % i)
            namespace = self.__get_tag_namespace(region_name % i)

            node = self.__add_section(PEOPLE)
            self.model.add([namespace, label, person_thumbnail, name], node=node)

            i += 1

    def __get_thumbnail(self, region, thumbnail_size):
        """
        Returns the thumbnail of the given region.
        """
        w = region[2] - region[0]
        h = region[3] - region[1]

        if (
            w <= self.pixbuf.get_width()
            and h <= self.pixbuf.get_height()
            and self.pixbuf
        ):
            subpixbuf = self.pixbuf.new_subpixbuf(region[0], region[1], w, h)
            size = self.__resize_keep_aspect(w, h, *thumbnail_size)
            return subpixbuf.scale_simple(
                size[0], size[1], GdkPixbuf.InterpType.BILINEAR
            )
        else:
            return None

    def __resize_keep_aspect(self, orig_x, orig_y, target_x, target_y):
        """
        Calculates the dimensions of the rectangle obtained from
        the rectangle orig_x * orig_y by scaling to fit
        target_x * target_y keeping the aspect ratio.
        """
        orig_aspect = orig_x / orig_y
        target_aspect = target_x / target_y
        if orig_aspect > target_aspect:
            return (target_x, target_x * orig_y // orig_x)
        else:
            return (target_y * orig_x // orig_y, target_y)

    def __get_tag_namespace(self, key):
        x = key.split(".")
        del x[-1]
        namespace = ".".join(x)

        return namespace



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2008  Brian G. Matherly
# Copyright (C) 2010  Jakim Friant
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
The "plug" package for handling plugins in Gramps.
"""

from ._plugin import Plugin
from ._pluginreg import (
    PluginData,
    PluginRegister,
    REPORT,
    TOOL,
    CATEGORY_TEXT,
    CATEGORY_DRAW,
    CATEGORY_CODE,
    CATEGORY_WEB,
    CATEGORY_BOOK,
    CATEGORY_GRAPHVIZ,
    CATEGORY_TREE,
    TOOL_DEBUG,
    TOOL_ANAL,
    TOOL_DBPROC,
    TOOL_DBFIX,
    TOOL_REVCTL,
    TOOL_UTILS,
    CATEGORY_QR_MISC,
    CATEGORY_QR_PERSON,
    CATEGORY_QR_FAMILY,
    CATEGORY_QR_EVENT,
    CATEGORY_QR_SOURCE,
    CATEGORY_QR_PLACE,
    CATEGORY_QR_REPOSITORY,
    CATEGORY_QR_NOTE,
    CATEGORY_QR_DATE,
    PTYPE_STR,
    CATEGORY_QR_MEDIA,
    CATEGORY_QR_CITATION,
    CATEGORY_QR_SOURCE_OR_CITATION,
    START,
    END,
    make_environment,
    AUDIENCETEXT,
    STATUSTEXT,
)
from ._import import ImportPlugin
from ._export import ExportPlugin
from ._docgenplugin import DocGenPlugin
from ._manager import BasePluginManager
from ._gramplet import Gramplet
from ._thumbnailer import Thumbnailer
from .utils import *
from ._options import (
    Options,
    OptionListCollection,
    OptionList,
    OptionHandler,
    MenuOptions,
)

__all__ = [
    "docbackend",
    "docgen",
    "menu",
    "Plugin",
    "PluginData",
    "PluginRegister",
    "BasePluginManager",
    "ImportPlugin",
    "ExportPlugin",
    "DocGenPlugin",
    "REPORT",
    "TOOL",
    "CATEGORY_TEXT",
    "CATEGORY_DRAW",
    "CATEGORY_CODE",
    "CATEGORY_WEB",
    "CATEGORY_BOOK",
    "CATEGORY_GRAPHVIZ",
    "CATEGORY_TREE",
    "TOOL_DEBUG",
    "TOOL_ANAL",
    "TOOL_DBPROC",
    "TOOL_DBFIX",
    "TOOL_REVCTL",
    "TOOL_UTILS",
    "CATEGORY_QR_MISC",
    "CATEGORY_QR_PERSON",
    "CATEGORY_QR_FAMILY",
    "CATEGORY_QR_EVENT",
    "CATEGORY_QR_SOURCE",
    "CATEGORY_QR_PLACE",
    "CATEGORY_QR_REPOSITORY",
    "CATEGORY_QR_NOTE",
    "CATEGORY_QR_DATE",
    "PTYPE_STR",
    "CATEGORY_QR_MEDIA",
    "CATEGORY_QR_CITATION",
    "CATEGORY_QR_SOURCE_OR_CITATION",
    "START",
    "END",
    "make_environment",
]



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2010       Michiel D. Nauta
# Copyright (C) 2011       Tim G L Lyons
# Copyright (C) 2017       Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Repository object for Gramps.
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..const import GRAMPS_LOCALE as glocale
from .addressbase import AddressBase
from .citationbase import IndirectCitationBase
from .notebase import NoteBase
from .primaryobj import PrimaryObject
from .repotype import RepositoryType
from .tagbase import TagBase
from .urlbase import UrlBase

_ = glocale.translation.gettext


# -------------------------------------------------------------------------
#
# Repository
#
# -------------------------------------------------------------------------
class Repository(NoteBase, AddressBase, UrlBase, IndirectCitationBase, PrimaryObject):
    """A location where collections of Sources are found."""

    def __init__(self):
        """
        Create a new Repository instance.
        """
        PrimaryObject.__init__(self)
        NoteBase.__init__(self)
        AddressBase.__init__(self)
        UrlBase.__init__(self)
        self.type = RepositoryType()
        self.name = ""

    def serialize(self):
        """
        Convert the object to a serialized tuple of data.
        """
        return (
            self.handle,
            self.gramps_id,
            self.type.serialize(),
            str(self.name),
            NoteBase.serialize(self),
            AddressBase.serialize(self),
            UrlBase.serialize(self),
            self.change,
            TagBase.serialize(self),
            self.private,
        )

    @classmethod
    def get_schema(cls):
        """
        Returns the JSON Schema for this class.

        :returns: Returns a dict containing the schema.
        :rtype: dict
        """
        # pylint: disable=import-outside-toplevel
        from .address import Address
        from .url import Url

        return {
            "type": "object",
            "title": _("Repository"),
            "properties": {
                "_class": {"enum": [cls.__name__]},
                "handle": {
                    "type": "string",
                    "maxLength": 50,
                    "title": _("Handle"),
                },
                "gramps_id": {"type": "string", "title": _("Gramps ID")},
                "type": RepositoryType.get_schema(),
                "name": {"type": "string", "title": _("Name")},
                "note_list": {
                    "type": "array",
                    "items": {"type": "string", "maxLength": 50},
                    "title": _("Notes"),
                },
                "address_list": {
                    "type": "array",
                    "items": Address.get_schema(),
                    "title": _("Addresses"),
                },
                "urls": {
                    "type": "array",
                    "items": Url.get_schema(),
                    "title": _("URLs"),
                },
                "change": {"type": "integer", "title": _("Last changed")},
                "tag_list": {
                    "type": "array",
                    "items": {"type": "string", "maxLength": 50},
                    "title": _("Tags"),
                },
                "private": {"type": "boolean", "title": _("Private")},
            },
        }

    def unserialize(self, data):
        """
        Convert the data held in a tuple created by the serialize method
        back into the data in a Repository structure.
        """
        (
            self.handle,
            self.gramps_id,
            the_type,
            self.name,
            note_list,
            address_list,
            urls,
            self.change,
            tag_list,
            self.private,
        ) = data

        self.type = RepositoryType()
        self.type.unserialize(the_type)
        NoteBase.unserialize(self, note_list)
        AddressBase.unserialize(self, address_list)
        UrlBase.unserialize(self, urls)
        TagBase.unserialize(self, tag_list)
        return self

    def get_text_data_list(self):
        """
        Return the list of all textual attributes of the object.

        :returns: Returns the list of all textual attributes of the object.
        :rtype: list
        """
        return [self.name, str(self.type), self.gramps_id]

    def get_text_data_child_list(self):
        """
        Return the list of child objects that may carry textual data.

        :returns: Returns the list of child objects that may carry textual data.
        :rtype: list
        """
        return self.address_list + self.urls

    def get_citation_child_list(self):
        """
        Return the list of child secondary objects that may refer citations.

        :returns: Returns the list of child secondary child objects that may
                  refer citations.
        :rtype: list
        """
        return self.address_list

    def get_note_child_list(self):
        """
        Return the list of child secondary objects that may refer notes.

        :returns: Returns the list of child secondary child objects that may
                  refer notes.
        :rtype: list
        """
        return self.address_list

    def get_handle_referents(self):
        """
        Return the list of child objects which may, directly or through
        their children, reference primary objects.

        :returns: Returns the list of objects referencing primary objects.
        :rtype: list
        """
        return self.get_citation_child_list()

    def get_referenced_handles(self):
        """
        Return the list of (classname, handle) tuples for all directly
        referenced primary objects.

        :returns: List of (classname, handle) tuples for referenced objects.
        :rtype: list
        """
        return self.get_referenced_note_handles() + self.get_referenced_tag_handles()

    def merge(self, acquisition):
        """
        Merge the content of acquisition into this repository.

        :param acquisition: The repository to merge with the present repository.
        :type acquisition: Repository
        """
        self._merge_privacy(acquisition)
        self._merge_address_list(acquisition)
        self._merge_url_list(acquisition)
        self._merge_note_list(acquisition)
        self._merge_tag_list(acquisition)

    def set_type(self, the_type):
        """
        :param the_type: descriptive type of the Repository
        :type the_type: str
        """
        self.type.set(the_type)

    def get_type(self):
        """
        :returns: the descriptive type of the Repository
        :rtype: str
        """
        return self.type

    def set_name(self, name):
        """
        :param name: descriptive name of the Repository
        :type name: str
        """
        self.name = name

    def get_name(self):
        """
        :returns: the descriptive name of the Repository
        :rtype: str
        """
        return self.name



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2010       Michiel D. Nauta
# Copyright (C) 2010,2017  Nick Hall
# Copyright (C) 2024       Doug Blank
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Note class for Gramps.
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..const import GRAMPS_LOCALE as glocale
from .notetype import NoteType
from .primaryobj import BasicPrimaryObject
from .styledtext import StyledText
from .styledtexttagtype import StyledTextTagType
from .tagbase import TagBase

_ = glocale.translation.gettext


# -------------------------------------------------------------------------
#
# Note
#
# -------------------------------------------------------------------------
class Note(BasicPrimaryObject):
    """Define a text note.

    Starting from Gramps 3.1 Note object stores the text in
    :class:`~.styledtext.StyledText` instance, thus it can have text formatting
    information.

    To get and set only the clear text of the note use the :meth:`get` and
    :meth:`set` methods.

    To get and set the formatted version of the Note's text use the
    :meth:`get_styledtext` and :meth:`set_styledtext` methods.

    The note may be 'preformatted' or 'flowed', which indicates that the
    text string is considered to be in paragraphs, separated by newlines.

    :cvar FLOWED: indicates flowed format
    :cvar FORMATTED: indicates formatted format (respecting whitespace needed)
    """

    (FLOWED, FORMATTED) = list(range(2))

    def __init__(self, text=""):
        """Create a new Note object, initializing from the passed string."""
        BasicPrimaryObject.__init__(self)
        self.text = StyledText(text)
        self.format = Note.FLOWED
        self.type = NoteType()

    def serialize(self):
        """Convert the object to a serialized tuple of data.

        :returns: The serialized format of the instance.
        :rtype: tuple

        """
        return (
            self.handle,
            self.gramps_id,
            self.text.serialize(),
            self.format,
            self.type.serialize(),
            self.change,
            TagBase.serialize(self),
            self.private,
        )

    @classmethod
    def get_schema(cls):
        """
        Returns the JSON Schema for this class.

        :returns: Returns a dict containing the schema.
        :rtype: dict
        """
        return {
            "type": "object",
            "title": _("Note"),
            "properties": {
                "_class": {"enum": [cls.__name__]},
                "handle": {
                    "type": "string",
                    "maxLength": 50,
                    "title": _("Handle"),
                },
                "gramps_id": {"type": "string", "title": _("Gramps ID")},
                "text": StyledText.get_schema(),
                "format": {"type": "integer", "title": _("Format")},
                "type": NoteType.get_schema(),
                "change": {"type": "integer", "title": _("Last changed")},
                "tag_list": {
                    "type": "array",
                    "items": {"type": "string", "maxLength": 50},
                    "title": _("Tags"),
                },
                "private": {"type": "boolean", "title": _("Private")},
            },
        }

    def unserialize(self, data):
        """Convert a serialized tuple of data to an object.

        :param data: The serialized format of a Note.
        :type: data: tuple
        """
        (
            self.handle,
            self.gramps_id,
            the_text,
            self.format,
            the_type,
            self.change,
            tag_list,
            self.private,
        ) = data

        self.text = StyledText()
        self.text.unserialize(the_text)
        self.type = NoteType()
        self.type.unserialize(the_type)
        TagBase.unserialize(self, tag_list)
        return self

    def get_text_data_list(self):
        """Return the list of all textual attributes of the object.

        :returns: The list of all textual attributes of the object.
        :rtype: list
        """
        return [str(self.text), self.gramps_id]

    def get_referenced_handles(self):
        """
        Return the list of (classname, handle) tuples for all directly
        referenced primary objects.

        :returns: List of (classname, handle) tuples for referenced objects.
        :rtype: list
        """
        reflist = []
        for dom, obj, prop, hndl in self.get_links():
            if dom != "gramps" or prop != "handle":
                continue
            reflist.append((obj, hndl))
        reflist.extend(self.get_referenced_tag_handles())
        return reflist

    def has_handle_reference(self, classname, handle):
        """
        Return True if the object has reference to a given handle of given
        primary object type.

        :param classname: The name of the primary object class.
        :type classname: str
        :param handle: The handle to be checked.
        :type handle: str

        :returns:
          Returns whether the object has reference to this handle of
          this object type.

        :rtype: bool
        """
        for dom, obj, prop, hndl in self.get_links():
            if (
                dom == "gramps"
                and prop == "handle"
                and obj == classname
                and hndl == handle
            ):
                return True
        return False

    def remove_handle_references(self, classname, handle_list):
        """
        Remove all references in this object to object handles in the list.

        :param classname: The name of the primary object class.
        :type classname: str
        :param handle_list: The list of handles to be removed.
        :type handle_list: str

        If the link is in the styled text, we just remove the style for that
        link.
        """
        tags = []
        for styledtext_tag in self.text.get_tags():
            if (
                styledtext_tag.name == StyledTextTagType.LINK
                and styledtext_tag.value.startswith("gramps://")
            ):
                obj, prop, value = styledtext_tag.value[9:].split("/", 2)
                if obj == classname and prop == "handle" and value in handle_list:
                    continue
            tags.append(styledtext_tag)
        self.text.set_tags(tags)

    def replace_handle_reference(self, classname, old_handle, new_handle):
        """
        Replace all references to old handle with those to the new handle.

        :param classname: The name of the primary object class.
        :type classname: str
        :param old_handle: The handle to be replaced.
        :type old_handle: str
        :param new_handle: The handle to replace the old one with.
        :type new_handle: str
        """
        for styledtext_tag in self.text.get_tags():
            if (
                styledtext_tag.name == StyledTextTagType.LINK
                and styledtext_tag.value.startswith("gramps://")
            ):
                obj, prop, value = styledtext_tag.value[9:].split("/", 2)
                if obj == classname and prop == "handle" and value == old_handle:
                    styledtext_tag.value = styledtext_tag.value.replace(
                        old_handle, new_handle
                    )

    def merge(self, acquisition):
        """
        Merge the content of acquisition into this note.

        Lost: handle, id, type, format, text and styles of acquisition.

        :param acquisition: The note to merge with the present note.
        :type acquisition: Note
        """
        self._merge_privacy(acquisition)
        self._merge_tag_list(acquisition)

    def set(self, text):
        """Set the text associated with the note to the passed string.

        :param text: The *clear* text defining the note contents.
        :type text: str
        """
        self.text = StyledText(text)

    def get(self):
        """Return the text string associated with the note.

        :returns: The *clear* text of the note contents.
        :rtype: unicode
        """
        return str(self.text)

    def set_styledtext(self, text):
        """Set the text associated with the note to the passed string.

        :param text: The *formatted* text defining the note contents.
        :type text: :class:`~.styledtext.StyledText`
        """
        self.text = text

    def get_styledtext(self):
        """Return the text string associated with the note.

        :returns: The *formatted* text of the note contents.
        :rtype: :class:`~.styledtext.StyledText`
        """
        return self.text

    def append(self, text):
        """Append the specified text to the text associated with the note.

        :param text: Text string to be appended to the note.
        :type text: str or :class:`~.styledtext.StyledText`
        """
        self.text = self.text + text

    def set_format(self, note_format):
        """Set the format of the note to the passed value.

        :param format: The value can either indicate Flowed or Preformatted.
        :type format: int
        """
        self.format = note_format

    def get_format(self):
        """Return the format of the note.

        The value can either indicate Flowed or Preformatted.

        :returns: 0 indicates Flowed, 1 indicates Preformated
        :rtype: int
        """
        return self.format

    def set_type(self, the_type):
        """Set descriptive type of the Note.

        :param the_type: descriptive type of the Note
        :type the_type: str
        """
        self.type.set(the_type)

    def get_type(self):
        """Get descriptive type of the Note.

        :returns: the descriptive type of the Note
        :rtype: str
        """
        return self.type

    def get_links(self):
        """
        Get the jump links from this note. Links can be external, to
        urls, or can be internal to gramps objects.

        Return examples::

            [("gramps", "Person", "handle", "7657626365362536"),
             ("external", "www", "url", "http://example.com")]

        :returns: list of [(domain, type, propery, value), ...]
        :rtype: list
        """
        retval = []
        for styledtext_tag in self.text.get_tags():
            if int(styledtext_tag.name) == StyledTextTagType.LINK:
                if styledtext_tag.value.startswith("gramps://"):
                    object_class, prop, value = styledtext_tag.value[9:].split("/", 2)
                    retval.append(("gramps", object_class, prop, value))
                else:
                    retval.append(("external", "www", "url", styledtext_tag.value))
        return retval



# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2011 Nick Hall
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# ------------------------------------------------------------------------
#
# Gtk
#
# ------------------------------------------------------------------------
from gi.repository import Gtk

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
from gramps.gen.plug import Gramplet
from gramps.gui.widgets import Photo
from gramps.gen.utils.thumbnails import get_thumbnail_image
from gramps.gen.utils.file import media_path_full


class Gallery(Gramplet):
    """
    Displays a gallery of media objects.
    """

    def init(self):
        self.gui.WIDGET = self.build_gui()
        self.gui.get_container_widget().remove(self.gui.textview)
        self.gui.get_container_widget().add(self.gui.WIDGET)

    def build_gui(self):
        """
        Build the GUI interface.
        """
        self.image_list = []
        self.top = Gtk.Box(spacing=3)
        return self.top

    def set_orientation(self, orientation):
        """
        Called when the gramplet orientation changes.
        """
        self.top.set_orientation(orientation)

    def clear_images(self):
        """
        Remove all images from the Gramplet.
        """
        for image in self.image_list:
            self.top.remove(image)
        self.image_list = []

    def load_images(self, obj):
        """
        Load the primary image into the main form if it exists.
        """
        media_list = obj.get_media_list()
        count = 0
        for media_ref in media_list:
            media_handle = media_ref.get_reference_handle()
            media = self.dbstate.db.get_media_from_handle(media_handle)
            full_path = media_path_full(self.dbstate.db, media.get_path())
            mime_type = media.get_mime_type()
            if mime_type and mime_type.startswith("image"):
                photo = Photo(self.uistate.screen_height() < 1000)
                photo.set_image(full_path, mime_type, media_ref.get_rectangle())
                photo.set_uistate(self.uistate, media_handle)
            else:
                photo = Photo(self.uistate.screen_height() < 1000)
                photo.set_pixbuf(
                    full_path,
                    get_thumbnail_image(
                        full_path, mime_type, media_ref.get_rectangle()
                    ),
                )
            self.image_list.append(photo)
            self.top.pack_start(photo, False, False, 0)
            count += 1
        self.top.show_all()
        self.set_has_data(count > 0)

    def get_has_data(self, obj):
        """
        Return True if the gramplet has data, else return False.
        """
        if obj is None:
            return False
        media_list = obj.get_media_list()
        for media_ref in media_list:
            media_handle = media_ref.get_reference_handle()
            media = self.dbstate.db.get_media_from_handle(media_handle)
            mime_type = media.get_mime_type()
            # bug 7390 : tab is not highlighted if there are only media
            #           like pdf, open document, ...
            if mime_type:
                return True
        return False


class PersonGallery(Gallery):
    """
    Displays a gallery of media objects for a person.
    """

    def db_changed(self):
        self.connect(self.dbstate.db, "person-update", self.update)

    def active_changed(self, handle):
        self.update()

    def update_has_data(self):
        active_handle = self.get_active("Person")
        if active_handle:
            active = self.dbstate.db.get_person_from_handle(active_handle)
            self.set_has_data(self.get_has_data(active))
        else:
            self.set_has_data(False)

    def main(self):
        self.clear_images()
        active_handle = self.get_active("Person")
        if active_handle:
            active = self.dbstate.db.get_person_from_handle(active_handle)
            if active:
                self.load_images(active)
            else:
                self.set_has_data(False)
        else:
            self.set_has_data(False)


class FamilyGallery(Gallery):
    """
    Displays a gallery of media objects for a family.
    """

    def db_changed(self):
        self.connect(self.dbstate.db, "family-update", self.update)
        self.connect_signal("Family", self.update)

    def update_has_data(self):
        active_handle = self.get_active("Family")
        if active_handle:
            active = self.dbstate.db.get_family_from_handle(active_handle)
            self.set_has_data(self.get_has_data(active))
        else:
            self.set_has_data(False)

    def main(self):
        self.clear_images()
        active_handle = self.get_active("Family")
        if active_handle:
            active = self.dbstate.db.get_family_from_handle(active_handle)
            if active:
                self.load_images(active)
            else:
                self.set_has_data(False)
        else:
            self.set_has_data(False)


class EventGallery(Gallery):
    """
    Displays a gallery of media objects for an event.
    """

    def db_changed(self):
        self.connect(self.dbstate.db, "event-update", self.update)
        self.connect_signal("Event", self.update)

    def update_has_data(self):
        active_handle = self.get_active("Event")
        if active_handle:
            active = self.dbstate.db.get_event_from_handle(active_handle)
            self.set_has_data(self.get_has_data(active))
        else:
            self.set_has_data(False)

    def main(self):
        self.clear_images()
        active_handle = self.get_active("Event")
        if active_handle:
            active = self.dbstate.db.get_event_from_handle(active_handle)
            if active:
                self.load_images(active)
            else:
                self.set_has_data(False)
        else:
            self.set_has_data(False)


class PlaceGallery(Gallery):
    """
    Displays a gallery of media objects for a place.
    """

    def db_changed(self):
        self.connect(self.dbstate.db, "place-update", self.update)
        self.connect_signal("Place", self.update)

    def update_has_data(self):
        active_handle = self.get_active("Place")
        if active_handle:
            active = self.dbstate.db.get_place_from_handle(active_handle)
            self.set_has_data(self.get_has_data(active))
        else:
            self.set_has_data(False)

    def main(self):
        self.clear_images()
        active_handle = self.get_active("Place")
        if active_handle:
            active = self.dbstate.db.get_place_from_handle(active_handle)
            if active:
                self.load_images(active)
            else:
                self.set_has_data(False)
        else:
            self.set_has_data(False)


class SourceGallery(Gallery):
    """
    Displays a gallery of media objects for a source.
    """

    def db_changed(self):
        self.connect(self.dbstate.db, "event-update", self.update)
        self.connect_signal("Source", self.update)

    def update_has_data(self):
        active_handle = self.get_active("Source")
        if active_handle:
            active = self.dbstate.db.get_source_from_handle(active_handle)
            self.set_has_data(self.get_has_data(active))
        else:
            self.set_has_data(False)

    def main(self):
        self.clear_images()
        active_handle = self.get_active("Source")
        if active_handle:
            active = self.dbstate.db.get_source_from_handle(active_handle)
            if active:
                self.load_images(active)
            else:
                self.set_has_data(False)
        else:
            self.set_has_data(False)


class CitationGallery(Gallery):
    """
    Displays a gallery of media objects for a Citation.
    """

    def db_changed(self):
        self.connect(self.dbstate.db, "event-update", self.update)
        self.connect_signal("Citation", self.update)

    def update_has_data(self):
        active_handle = self.get_active("Citation")
        if active_handle:
            active = self.dbstate.db.get_citation_from_handle(active_handle)
            self.set_has_data(self.get_has_data(active))
        else:
            self.set_has_data(False)

    def main(self):
        self.clear_images()
        active_handle = self.get_active("Citation")
        if active_handle:
            active = self.dbstate.db.get_citation_from_handle(active_handle)
            if active:
                self.load_images(active)
            else:
                self.set_has_data(False)
        else:
            self.set_has_data(False)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2011      Michiel D. Nauta
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Unittest for export to VCard
"""
import unittest
import time
import subprocess
import sys
import os
import xml.etree.ElementTree as ET

from ...lib.libgrampsxml import GRAMPS_XML_VERSION
from gramps.version import VERSION
from ..exportvcard import VCardWriter


class VCardCheck(unittest.TestCase):
    def setUp(self):
        self.expect = [
            "BEGIN:VCARD",
            "VERSION:3.0",
            "PRODID:-//Gramps//NONSGML Gramps %s//EN" % VERSION,
            "FN:Lastname",
            "N:Lastname;;;;",
            "SORT-STRING:" + "Lastname".ljust(55),
            "END:VCARD",
        ]
        date = time.localtime(time.time())
        self.input_list = [
            "BEGIN:VCARD",
            "VERSION:3.0",
            "FN:Lastname",
            "N:Lastname;;;;",
            "END:VCARD",
        ]
        self.header = """<?xml version="1.0" encoding="UTF-8"?>
            <!DOCTYPE database PUBLIC "-//GRAMPS//DTD GRAMPS XML %s//EN"
            "http://gramps-project.org/xml/%s/grampsxml.dtd">""" % (
            GRAMPS_XML_VERSION,
            GRAMPS_XML_VERSION,
        )
        strng = """<database xmlns="http://gramps-project.org/xml/%s/">
            <header>
            <created date="%04d-%02d-%02d" version="%s"/>
            <researcher/>
            </header>
            <people>
            <person id="I0000" handle="_0000">
            <name type="Birth Name">
            <surname>Lastname</surname>
            </name>
            </person>
            </people>
            </database>""" % (
            GRAMPS_XML_VERSION,
            date[0],
            date[1],
            date[2],
            VERSION,
        )
        namespace = "http://gramps-project.org/xml/%s/" % GRAMPS_XML_VERSION
        ET.register_namespace("", namespace)
        self.database = ET.XML(strng)
        self.people = self.database[1]
        self.person = self.people[0]
        self.name = self.person[0]
        self.lastname = self.name[0]

    def do_case(self, input_doc, expect_str, debug=False):
        if debug:
            print(ET.tostring(input_doc))

        gcmd = [
            sys.executable,
            "Gramps.py",
            "-i",
            "-",
            "-f",
            "gramps",
            "-e",
            "-",
            "-f",
            "vcf",
        ]
        process = subprocess.Popen(
            gcmd,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            env=os.environ,
        )
        input_str = self.header.encode("utf-8") + ET.tostring(
            input_doc, encoding="utf-8"
        )
        result_str, err_str = process.communicate(input_str)

        separator = "\r" + os.linesep
        expect_str = separator.join(expect_str) + (separator * 2)

        if debug:
            print(err_str)
            print(result_str)
            print(expect_str)
        self.assertEqual(result_str, expect_str.encode("utf-8"))

    def test_base(self):
        self.do_case(self.database, self.expect)

    def test_esc_string_none(self):
        self.assertEqual(VCardWriter.esc("nothing"), "nothing")

    def test_esc_string_all(self):
        self.assertEqual(
            VCardWriter.esc("backslash\\_comma,_semicolon;"),
            "backslash\\\\_comma\\,_semicolon\\;",
        )

    def test_esc_string_list(self):
        self.assertEqual(
            VCardWriter.esc(["comma,", "semicolon;"]), ["comma\\,", "semicolon\\;"]
        )

    def test_esc_string_tuple(self):
        self.assertEqual(
            VCardWriter.esc(("comma,", "semicolon;")), ("comma\\,", "semicolon\\;")
        )

    def test_esc_string_wrongtype(self):
        self.assertRaises(TypeError, VCardWriter.esc, {"comma,": "semicolon;"})

    def test_write_formatted_name_title(self):
        ET.SubElement(self.name, "title").text = "Sir."
        self.expect[3] = "FN:Sir. Lastname"
        self.expect[4] = "N:Lastname;;;Sir.;"
        self.do_case(self.database, self.expect)

    def test_write_name_multiple_surname(self):
        self.lastname.text = "Oranje"
        self.lastname.set("prefix", "van")
        ET.SubElement(self.name, "surname").text = "Nassau"
        self.expect[3] = "FN:van Oranje Nassau"
        self.expect[4] = "N:van Oranje,Nassau;;;;"
        self.expect[5] = "SORT-STRING:" + "Oranje".ljust(55)
        self.do_case(self.database, self.expect)

    def test_write_name_callname(self):
        # callname not in first names!
        ET.SubElement(self.name, "first").text = "B C"
        ET.SubElement(self.name, "call").text = "A"
        self.expect[3] = "FN:B C Lastname"
        self.expect[4] = "N:Lastname;A;B,C;;"
        self.expect[5] = "SORT-STRING:" + "Lastname".ljust(25) + "B C".ljust(30)
        self.do_case(self.database, self.expect)

    def test_write_name_callname_in_addnames(self):
        ET.SubElement(self.name, "first").text = "A B C"
        ET.SubElement(self.name, "call").text = "B"
        self.expect[3] = "FN:A B C Lastname"
        self.expect[4] = "N:Lastname;B;A,C;;"
        self.expect[5] = "SORT-STRING:" + "Lastname".ljust(25) + "A B C".ljust(30)
        self.do_case(self.database, self.expect)

    def test_write_name_no_callname(self):
        ET.SubElement(self.name, "first").text = "A B C"
        self.expect[3] = "FN:A B C Lastname"
        self.expect[4] = "N:Lastname;A;B,C;;"
        self.expect[5] = "SORT-STRING:" + "Lastname".ljust(25) + "A B C".ljust(30)
        self.do_case(self.database, self.expect)

    def test_write_name_no_additional_names(self):
        ET.SubElement(self.name, "first").text = "A"
        self.expect[3] = "FN:A Lastname"
        self.expect[4] = "N:Lastname;A;;;"
        self.expect[5] = "SORT-STRING:" + "Lastname".ljust(25) + "A".ljust(30)
        self.do_case(self.database, self.expect)

    def test_write_name_honprefix(self):
        ET.SubElement(self.name, "title").text = "Sir"
        self.expect[3] = "FN:Sir Lastname"
        self.expect[4] = "N:Lastname;;;Sir;"
        self.expect[5] = "SORT-STRING:" + "Lastname".ljust(55)
        self.do_case(self.database, self.expect)

    def test_write_name_honsuffix(self):
        ET.SubElement(self.name, "suffix").text = "Jr."
        self.expect[3] = "FN:Lastname\\, Jr."
        self.expect[4] = "N:Lastname;;;;Jr."
        self.expect[5] = "SORT-STRING:" + "Lastname".ljust(55) + "Jr."
        self.do_case(self.database, self.expect)

    def test_nicknames_regular(self):
        attribs = {"type": "Birth Name", "alt": "1"}
        name = ET.SubElement(self.person, "name", attrib=attribs)
        ET.SubElement(name, "nick").text = "Nick"
        name = ET.SubElement(self.person, "name", attrib=attribs)
        ET.SubElement(name, "nick").text = "N."
        self.expect.insert(6, "NICKNAME:Nick,N.")
        self.do_case(self.database, self.expect)

    def test_nicknames_primary_nick(self):
        ET.SubElement(self.name, "nick").text = "Nick"
        attribs = {"type": "Birth Name", "alt": "1"}
        name = ET.SubElement(self.person, "name", attrib=attribs)
        ET.SubElement(name, "nick").text = "N."
        self.expect.insert(6, "NICKNAME:Nick,N.")
        self.do_case(self.database, self.expect)

    def test_write_birthdate_regular(self):
        events = ET.Element("events")
        self.database.insert(1, events)
        attribs = {"handle": "_e0000", "id": "E0000"}
        event = ET.SubElement(events, "event", attrib=attribs)
        ET.SubElement(event, "type").text = "Birth"
        ET.SubElement(event, "dateval", val="2001-02-28")
        attribs = {"hlink": "_e0000", "role": "Primary"}
        ET.SubElement(self.person, "eventref", attrib=attribs)
        self.expect.insert(6, "BDAY:2001-02-28")
        self.do_case(self.database, self.expect)

    def test_write_birthdate_empty(self):
        events = ET.Element("events")
        self.database.insert(1, events)
        attribs = {"handle": "_e0000", "id": "E0000"}
        event = ET.SubElement(events, "event", attrib=attribs)
        ET.SubElement(event, "type").text = "Birth"
        attribs = {"hlink": "_e0000", "role": "Primary"}
        ET.SubElement(self.person, "eventref", attrib=attribs)
        self.do_case(self.database, self.expect)

    def test_write_birhtdate_textonly(self):
        events = ET.Element("events")
        self.database.insert(1, events)
        attribs = {"handle": "_e0000", "id": "E0000"}
        event = ET.SubElement(events, "event", attrib=attribs)
        ET.SubElement(event, "type").text = "Birth"
        ET.SubElement(event, "dateval", val="Christmas 2001")
        attribs = {"hlink": "_e0000", "role": "Primary"}
        ET.SubElement(self.person, "eventref", attrib=attribs)
        self.do_case(self.database, self.expect)

    def test_write_birthdate_span(self):
        events = ET.Element("events")
        self.database.insert(1, events)
        attribs = {"handle": "_e0000", "id": "E0000"}
        event = ET.SubElement(events, "event", attrib=attribs)
        ET.SubElement(event, "type").text = "Birth"
        attribs = {"start": "2001-02-28", "stop": "2002-02-28"}
        ET.SubElement(event, "datespan", attrib=attribs)
        attribs = {"hlink": "_e0000", "role": "Primary"}
        ET.SubElement(self.person, "eventref", attrib=attribs)
        self.do_case(self.database, self.expect)

    def test_write_birthdate_range(self):
        events = ET.Element("events")
        self.database.insert(1, events)
        attribs = {"handle": "_e0000", "id": "E0000"}
        event = ET.SubElement(events, "event", attrib=attribs)
        ET.SubElement(event, "type").text = "Birth"
        attribs = {"start": "2001-02-28", "stop": "2002-02-28"}
        ET.SubElement(event, "daterange", attrib=attribs)
        attribs = {"hlink": "_e0000", "role": "Primary"}
        ET.SubElement(self.person, "eventref", attrib=attribs)
        self.do_case(self.database, self.expect)

    def test_write_addresses_regular(self):
        address = ET.SubElement(self.person, "address")
        ET.SubElement(address, "street").text = "pobox bis street"
        ET.SubElement(address, "city").text = "place"
        ET.SubElement(address, "country").text = "country"
        ET.SubElement(address, "state").text = "province"
        ET.SubElement(address, "postal").text = "zip"
        self.expect.insert(6, "ADR:;;pobox bis street;place;province;zip;country")
        self.do_case(self.database, self.expect)

    def test_write_addresses_phone(self):
        address = ET.SubElement(self.person, "address")
        ET.SubElement(address, "phone").text = "01234-56789"
        self.expect.insert(6, "TEL:01234-56789")
        self.do_case(self.database, self.expect)

    def test_write_urls_email(self):
        attribs = {"type": "E-mail", "href": "me@example.com"}
        ET.SubElement(self.person, "url", attrib=attribs)
        self.expect.insert(6, "EMAIL:me@example.com")
        self.do_case(self.database, self.expect)

    def test_write_urls_emial_mailto(self):
        attribs = {"type": "E-mail", "href": "mailto:me@example.com"}
        ET.SubElement(self.person, "url", attrib=attribs)
        self.expect.insert(6, "EMAIL:me@example.com")
        self.do_case(self.database, self.expect)

    def test_write_urls_url(self):
        attribs = {"type": "Web Home", "href": "http://www.example.org"}
        ET.SubElement(self.person, "url", attrib=attribs)
        self.expect.insert(6, "URL:http://www.example.org")
        self.do_case(self.database, self.expect)

    def test_write_occupation_regular(self):
        events = ET.Element("events")
        self.database.insert(1, events)
        attribs = {"handle": "_e0000", "id": "E0000"}
        event = ET.SubElement(events, "event", attrib=attribs)
        ET.SubElement(event, "type").text = "Occupation"
        ET.SubElement(event, "description").text = "carpenter"
        attribs = {"hlink": "_e0000", "role": "Primary"}
        ET.SubElement(self.person, "eventref", attrib=attribs)
        self.expect.insert(6, "ROLE:carpenter")
        self.do_case(self.database, self.expect)

    def test_write_occupation_lastdate(self):
        events = ET.Element("events")
        self.database.insert(1, events)
        attribs = {"handle": "_e0000", "id": "E0000"}
        event = ET.SubElement(events, "event", attrib=attribs)
        ET.SubElement(event, "type").text = "Occupation"
        ET.SubElement(event, "dateval", val="2011-02-28")
        ET.SubElement(event, "description").text = "foreman"
        attribs = {"handle": "_e0001", "id": "E0001"}
        event = ET.SubElement(events, "event", attrib=attribs)
        ET.SubElement(event, "type").text = "Occupation"
        ET.SubElement(event, "dateval", val="2000-09-21")
        ET.SubElement(event, "description").text = "carpenter"
        attribs = {"hlink": "_e0000", "role": "Primary"}
        ET.SubElement(self.person, "eventref", attrib=attribs)
        attribs = {"hlink": "_e0001", "role": "Primary"}
        ET.SubElement(self.person, "eventref", attrib=attribs)
        self.expect.insert(6, "ROLE:foreman")
        self.do_case(self.database, self.expect)


if __name__ == "__main__":
    unittest.main()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2009       Gary Burton
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Make an 'Unknown' primary object
"""

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
from time import strftime, localtime, time
import os

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..lib import (
    Person,
    Surname,
    Name,
    NameType,
    Family,
    FamilyRelType,
    Event,
    EventType,
    Source,
    Place,
    Citation,
    Repository,
    RepositoryType,
    Media,
    Note,
    NoteType,
    StyledText,
    StyledTextTag,
    StyledTextTagType,
    Tag,
    ChildRef,
    ChildRefType,
)
from .id import create_id
from ..const import IMAGE_DIR
from ..const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext


# -------------------------------------------------------------------------
#
# make_unknown
#
# -------------------------------------------------------------------------
def make_unknown(class_arg, explanation, class_func, commit_func, transaction, **argv):
    """
    Make a primary object and set some property so that it qualifies as
    "Unknown".

    Some object types need extra parameters:
    Family: db, Event: type (optional),
    Citation: methods to create/store source.

    Some theoretical underpinning
    This function exploits the fact that all import methods basically do the
    same thing: Create an object of the right type, fill it with some
    attributes, store it in the database. This function does the same, so
    the observation is why not use the creation and storage methods that the
    import routines use themselves, that makes nice reuse of code. To do this
    formally correct we would need to specify a interface (in the OOP sence)
    which the import methods would need to implement. For now, that is deemed
    too restrictive and here we just slip through because of the similarity in
    code of both GEDCOM and XML import methods.

    :param class_arg: The argument the class_func needs, typically a kind of id.
    :type class_arg: unspecified
    :param explanation: Handle of a note that explains the origin of primary obj
    :type explanation: str
    :param class_func: Method to create primary object.
    :type class_func: method
    :param commit_func: Method to store primary object in db.
    :type commit_func: method
    :param transactino: Database transaction handle
    :type transaction: str
    :param argv: Possible additional parameters
    :type param: unspecified
    :returns: List of newly created objects.
    :rtype: list
    """
    retval = []
    obj = class_func(class_arg)
    if isinstance(obj, Person):
        surname = Surname()
        surname.set_surname("Unknown")
        name = Name()
        name.add_surname(surname)
        name.set_type(NameType.UNKNOWN)
        obj.set_primary_name(name)
    elif isinstance(obj, Family):
        obj.set_relationship(FamilyRelType.UNKNOWN)
        handle = obj.handle
        backlinks = argv["db"].find_backlink_handles(handle, [Person.__name__])
        for dummy, person_handle in backlinks:
            person = argv["db"].get_person_from_handle(person_handle)
            add_personref_to_family(obj, person)
    elif isinstance(obj, Event):
        if "type" in argv:
            obj.set_type(argv["type"])
        else:
            obj.set_type(EventType.UNKNOWN)
    elif isinstance(obj, Place):
        obj.set_title(_("Unknown"))
        obj.name.set_value(_("Unknown"))
    elif isinstance(obj, Source):
        obj.set_title(_("Unknown"))
    elif isinstance(obj, Citation):
        # TODO create a new source for every citation?
        obj2 = argv["source_class_func"](argv["source_class_arg"])
        obj2.set_title(_("Unknown"))
        obj2.add_note(explanation)
        argv["source_commit_func"](obj2, transaction, time())
        retval.append(obj2)
        obj.set_reference_handle(obj2.handle)
    elif isinstance(obj, Repository):
        obj.set_name(_("Unknown"))
        obj.set_type(RepositoryType.UNKNOWN)
    elif isinstance(obj, Media):
        obj.set_path("image-missing.png")
        obj.set_mime_type("image/png")
        obj.set_description(_("Unknown"))
    elif isinstance(obj, Note):
        obj.set_type(NoteType.UNKNOWN)
        text = _("Unknown, created to replace a missing note object.")
        link_start = text.index(",") + 2
        link_end = len(text) - 1
        tag = StyledTextTag(
            StyledTextTagType.LINK,
            "gramps://Note/handle/%s" % explanation,
            [(link_start, link_end)],
        )
        obj.set_styledtext(StyledText(text, [tag]))
    elif isinstance(obj, Tag):
        if not hasattr(make_unknown, "count"):
            make_unknown.count = 1  # primitive static variable
        obj.set_name(
            _("Unknown, was missing %(time)s (%(count)d)")
            % {"time": strftime("%x %X", localtime()), "count": make_unknown.count}
        )
        make_unknown.count += 1
    else:
        raise TypeError("Object if of unsupported type")

    if hasattr(obj, "add_note"):
        obj.add_note(explanation)
    commit_func(obj, transaction, time())
    retval.append(obj)
    return retval


def create_explanation_note(dbase):
    """
    When creating objects to fill missing primary objects in imported files,
    those objects of type "Unknown" need a explanatory note. This funcion
    provides such a note for import methods.
    """
    note = Note(
        _("Objects referenced by this note " "were missing in a file imported on %s.")
        % strftime("%x %X", localtime())
    )
    note.set_handle(create_id())
    note.set_gramps_id(dbase.find_next_note_gramps_id())
    # Use defaults for privacy, format and type.
    return note


def add_personref_to_family(family, person):
    """
    Given a family and person, set the parent/child references in the family,
    that match the person.
    """
    handle = family.handle
    person_handle = person.handle
    if handle in person.get_family_handle_list():
        if (person.get_gender() == Person.FEMALE) and (
            family.get_mother_handle() is None
        ):
            family.set_mother_handle(person_handle)
        else:
            # This includes cases of Person.UNKNOWN and Person.OTHER
            if family.get_father_handle() is None:
                family.set_father_handle(person_handle)
            else:
                family.set_mother_handle(person_handle)
    if handle in person.get_parent_family_handle_list():
        childref = ChildRef()
        childref.set_reference_handle(person_handle)
        childref.set_mother_relation(ChildRefType.UNKNOWN)
        childref.set_father_relation(ChildRefType.UNKNOWN)
        family.add_child_ref(childref)
